//*****************************************************************************
//*
//*
//*     SmbIpc.cpp
//*
//*
//*****************************************************************************
//
//  Copyright  2015    Anton Zechner
//
//  AzSmb is distributed under the GNU GENERAL PUBLIC LICENSE (GPL)
//  Sourcecode which use AzSmb must be published. Commercial users
//  must published their code too, or make an licence agreement with me.
//
//
//  AzSmb wird unter GNU GENERAL PUBLIC LICENSE (GPL) vertreiben.
//  Sourcecode welcher AzSmb verwendet muss verffentlicht werden.
//  Kommerzielle Nutzer mssen ihren Code ebenfalls verffentlichen, oder
//  eine Nutzungsvereinbarung mit mir treffen.
//
//  az_software@inode.at
//
//
//	This code is not a full implementation of the pipe access.
//	Is simulates only some responses for a fast access to the server.
//

#include	"System/System.h"
#include	"SmbInclude.h"
#include	"SmbInline.h"
#include	"SmbServer.h"
#include	"SmbMisc.h"
#include	"SmbIpc.h"
#include	"Smb.h"


#ifndef 	IPC_BUFFER_SIZE
#define		IPC_BUFFER_SIZE		MIN(1024,MAX_TREE_ENTRIES*128)			
#endif 	
#ifndef 	IPC_BUFF_WAIT
#define		IPC_BUFF_WAIT		5000			
#endif 	
#ifndef 	IPC_TIMEOUT
#define		IPC_TIMEOUT			3000			
#endif 	

#define     STYPE_DISKTREE              		0               // Disk drive
#define     STYPE_PRINTQ                		1               // Spooler queue
#define     STYPE_DEVICE						2               // Serial device
#define     STYPE_IPC							3               // Interprocess communication (IPC)
#define     STYPE_HIDDEN						0x80000000      // share is a hidden one (ends with $)

#define		PERM_ACCESS_READ					0x01
#define		PERM_ACCESS_WRITE        			0x02
#define		PERM_ACCESS_CREATE       			0x04
#define		PERM_ACCESS_EXEC         			0x08
#define		PERM_ACCESS_DELETE       			0x10
#define		PERM_ACCESS_ATRIB        			0x20
#define		PERM_ACCESS_PERM         			0x40

#define		DCERPC_PKT_REQUEST					0
#define		DCERPC_PKT_PING						1
#define		DCERPC_PKT_RESPONSE					2
#define		DCERPC_PKT_FAULT					3
#define		DCERPC_PKT_WORKING					4
#define		DCERPC_PKT_NOCALL					5
#define		DCERPC_PKT_REJECT					6
#define		DCERPC_PKT_ACK						7
#define		DCERPC_PKT_CL_CANCEL				8
#define		DCERPC_PKT_FACK						9
#define		DCERPC_PKT_CANCEL_ACK				10
#define		DCERPC_PKT_BIND						11
#define		DCERPC_PKT_BIND_ACK					12
#define		DCERPC_PKT_BIND_NAK					13
#define		DCERPC_PKT_ALTER					14
#define		DCERPC_PKT_ALTER_RESP				15
#define		DCERPC_PKT_AUTH3					16
#define		DCERPC_PKT_SHUTDOWN					17
#define		DCERPC_PKT_CO_CANCEL				18
#define		DCERPC_PKT_ORPHANED					19

#define 	SRVSVC_NETCHARDEVENUM 				0x00
#define 	SRVSVC_NETCHARDEVGETINFO 			0x01
#define 	SRVSVC_NETCHARDEVCONTROL 			0x02
#define 	SRVSVC_NETCHARDEVQENUM 				0x03
#define 	SRVSVC_NETCHARDEVQGETINFO 			0x04
#define 	SRVSVC_NETCHARDEVQSETINFO 			0x05
#define 	SRVSVC_NETCHARDEVQPURGE 			0x06
#define 	SRVSVC_NETCHARDEVQPURGESELF 		0x07
#define 	SRVSVC_NETCONNENUM 					0x08
#define 	SRVSVC_NETFILEENUM 					0x09
#define 	SRVSVC_NETFILEGETINFO 				0x0A
#define 	SRVSVC_NETFILECLOSE 				0x0B
#define 	SRVSVC_NETSESSENUM 					0x0C
#define 	SRVSVC_NETSESSDEL 					0x0D
#define 	SRVSVC_NETSHAREADD 					0x0E
#define 	SRVSVC_NETSHAREENUMALL 				0x0F
#define 	SRVSVC_NETSHAREGETINFO 				0x10
#define 	SRVSVC_NETSHARESETINFO 				0x11
#define 	SRVSVC_NETSHAREDEL 					0x12
#define 	SRVSVC_NETSHAREDELSTICKY 			0x13
#define 	SRVSVC_NETSHARECHECK 				0x14
#define 	SRVSVC_NETSRVGETINFO 				0x15
#define 	SRVSVC_NETSRVSETINFO 				0x16
#define 	SRVSVC_NETDISKENUM 					0x17
#define 	SRVSVC_NETSERVERSTATISTICSGET 		0x18
#define 	SRVSVC_NETTRANSPORTADD 				0x19
#define 	SRVSVC_NETTRANSPORTENUM 			0x1A
#define 	SRVSVC_NETTRANSPORTDEL 				0x1B
#define 	SRVSVC_NETREMOTETOD 				0x1C
#define 	SRVSVC_NETSETSERVICEBITS 			0x1D
#define 	SRVSVC_NETPATHTYPE 					0x1E
#define 	SRVSVC_NETPATHCANONICALIZE 			0x1F
#define 	SRVSVC_NETPATHCOMPARE 				0x20
#define 	SRVSVC_NETNAMEVALIDATE 				0x21
#define 	SRVSVC_NETRPRNAMECANONICALIZE 		0x22
#define 	SRVSVC_NETPRNAMECOMPARE 			0x23
#define 	SRVSVC_NETSHAREENUM 				0x24
#define 	SRVSVC_NETSHAREDELSTART 			0x25
#define 	SRVSVC_NETSHAREDELCOMMIT 			0x26
#define 	SRVSVC_NETGETFILESECURITY 			0x27
#define 	SRVSVC_NETSETFILESECURITY 			0x28
#define 	SRVSVC_NETSERVERTRANSPORTADDEX 		0x29
#define 	SRVSVC_NETSERVERSETSERVICEBITSEX 	0x2A
#define 	SRVSVC_NETRDFSGETVERSION 			0x2B
#define 	SRVSVC_NETRDFSCREATELOCALPARTITION 	0x2C
#define 	SRVSVC_NETRDFSDELETELOCALPARTITION 	0x2D
#define 	SRVSVC_NETRDFSSETLOCALVOLUMESTATE 	0x2E
#define 	SRVSVC_NETRDFSSETSERVERINFO 		0x2F
#define 	SRVSVC_NETRDFSCREATEEXITPOINT 		0x30
#define 	SRVSVC_NETRDFSDELETEEXITPOINT 		0x31
#define 	SRVSVC_NETRDFSMODIFYPREFIX 			0x32
#define 	SRVSVC_NETRDFSFIXLOCALVOLUME 		0x33
#define 	SRVSVC_NETRDFSMANAGERREPORTSITEINFO 0x34
#define 	SRVSVC_NETRSERVERTRANSPORTDELEX 	0x35


#define		WERR_OK								0x00000000
#define 	WERR_ACCESS_DENIED					0x00000005
#define 	WERR_BAD_NETPATH					0x00000035
#define		WERR_INVALID_PARAM					0x00000057
#define		WERR_CALL_NOT_IMPLEMENTED			0x00000078
#define		WERR_UNKNOWN_LEVEL					0x0000007C
			
#define 	v				volatile

static		SysSemaphore	hIpcSem;					// common ipc semaphore
static		SysSemaphore	hIpcLock;					// ipc buffer lock semaphore
static 		SmbFileEntry	sSrvSvcItem;				// file item for a tree
static		SmbPipe			sSrvSvcPipe;				// pipe entry for "/SRVSVC"
static		unsigned char	aIpcBuffer[IPC_BUFFER_SIZE];// buffer ipc data
static		unsigned short	wIpcBufferTid;				// tree id for lock
static		unsigned short	wIpcBufferFid;				// file id for lock
static		unsigned short	wIpcBufferMpx;				// muti id for lock
static		unsigned v		uIpcBufferSession;			// session id for lock
static		unsigned v		uIpcBufferMode;				// currend mode (0=free 1=write 2=read)
static		unsigned v 		uIpcBufferTick;				// start tick from locking
static		unsigned 		uIpcBufferWr;				// write position for puffer
static		unsigned 		uIpcBufferRd;				// read  position for puffer

typedef struct
	{
	unsigned char			ucVersionMajor;				// major version number (5)
	unsigned char			ucVersionMinor;				// minor version number (0)
	unsigned char			ucPacketType;				// 0x00=request 0x02=response
	unsigned char			ucPacketFlags;				// 0x01=first 0x02=last
	unsigned				uRepresentation;			// 0x10=littel endian
	unsigned short			usFragLength;				// fragment length (size of full, header+data)
	unsigned short			usAuthLength;				// ??? length
	unsigned				uCallId;					// id from the call
	}DecRpcHeader;

typedef struct
	{
	DecRpcHeader			sHeader;					// is the header
	unsigned				uAllocHint;					// allocation hint (size of packet without header)
	unsigned short			usContextId;				// context number
	unsigned short			usOperator;					// number of command
	}DecRpcRequest;	

typedef struct
	{
	DecRpcHeader			sHeader;					// is the header
	unsigned				uAllocHint;					// allocation hint (size of packet without header)
	unsigned short			usContextId;				// context number (take from request)
	unsigned char 			ucCancelCount;				// cancel count (0)
	unsigned char 			ucReserved;					// reserved (0)
	}DecRpcResponse;	

typedef struct
	{
	unsigned				uStatus;					// NT error number
	unsigned				uReserved;					// reserved 0x00000000 
	}DecRpcFaultResp;

typedef struct
	{
	unsigned				uMax;						// max string size
	unsigned				uOff;						// offset of string
	unsigned				uNum;						// size of string
	}DecStringW;


static int SrvScvWrite(SmbSession *pSession,SmbPipeEntryStr *pPipe,const void *pData,int iLength);
static int SrvScvRead (SmbSession *pSession,SmbPipeEntryStr *pPipe,      void *pData,int iLength);


//*****************************************************************************
//*
//*		SmbIpcInit
//*
//*****************************************************************************
//	initalizes the ipc features
//	pSession		: is the pointer to the smb session
//	returns 1 if ok or 0 if an error occurs.
int SmbIpcInit(SmbSession *pSession)
{

	if(!hIpcSem )hIpcSem =SysSemaphoreCreate();			// common ipc semaphore
	if(!hIpcLock)hIpcLock=SysSemaphoreCreate();			// ipc buffer lock semaphore

	sSrvSvcItem.bType		= 1;
	sSrvSvcItem.pPipe		= &sSrvSvcPipe;
	sSrvSvcPipe.pWriteCb	= SrvScvWrite;
	sSrvSvcPipe.pReadCb		= SrvScvRead;


return 1;
}

//*****************************************************************************
//*
//*		SmbIpcExit
//*
//*****************************************************************************
//	deinitalizes the ipc features
//	pSession		: is the pointer to the smb session
//	iTreeId			: is the colsed tree id or -1 for the full session
//	returns 1 if ok or 0 if an error occurs.
int SmbIpcExit(SmbSession *pSession,int iTreeId)
{

	
return 1;
}


//*****************************************************************************
//*
//*		SmbIpcLock
//*
//*****************************************************************************
//	Locks the ipc buffer for an access
//	pSession		: is the pointer to the smb session
//	iTreeId			: is the colsed tree id or -1 for the full session
//	Returns 1 if the buffer was locked or 0 if an error occurs.
//	Call SysSemaphoreUnlock(hIpcSem) if 1 was retuned
static int SmbIpcLock(SmbSession *pSession,unsigned uTimeout)
{
unsigned	uFlag;
unsigned	uStop;
unsigned	uDiff;
unsigned	uTurn;
unsigned	uTick;
int			iWait;

	
	
	uStop = SysTickCount()+uTimeout;
	uFlag = 0;


	for(uTurn=0;;uTurn++)
		{
		SysSemaphoreLock(hIpcSem);

		if(uIpcBufferMode==0)									// is the ipc buffer unused
			{
			if(SysSemaphoreLockWait(hIpcLock,50))
				{  
				uFlag = 1;
				break;
				}
			}
		else if(uIpcBufferSession==pSession->uId)				// is the request from the same session
			{
			if(wIpcBufferTid==pSession->wTid)
			if(wIpcBufferFid==pSession->wFid)
			if(wIpcBufferMpx==pSession->wMpx)
				{
				break;
				}
			
			SysSemaphoreUnlock(hIpcSem);
			
			return 0;
			}

		uTick	= SysTickCount();
		uDiff	= uTick-uIpcBufferTick;
		
		if(uDiff>IPC_TIMEOUT)									// buffer too long locked from other pipe
			{
			uFlag = 1;
			break;
			}

		SysSemaphoreUnlock(hIpcSem);


		   iWait = uStop-uTick;
		if(iWait<0)
			{
			if(uTurn>=2)return 0;
			iWait = 100;			
			} 


		if(SysSemaphoreLockWait(hIpcLock,iWait))				// wait until the pipe buffer goes free		
			{
			SysSemaphoreUnlock(hIpcLock);
			}
		}


	if(uFlag)													// take the control over the buffer
		{			
		uIpcBufferSession	= pSession->uId;
		uIpcBufferTick		= SysTickCount();
		wIpcBufferTid		= pSession->wTid; 
		wIpcBufferFid		= pSession->wFid;
		wIpcBufferMpx		= pSession->wMpx;
		uIpcBufferMode		= 1;
		uIpcBufferWr		= 0;
		}



return 1;
}




//*****************************************************************************
//*
//*		SmbGetPipe
//*
//*****************************************************************************
//	Gets a pointer to a pipe handle
//	usPipeId	: is the id of the pipe
//	Returns 
SmbFileEntry *SmbGetPipe(SmbSession *pSession,unsigned short usPipeId)
{

	switch(usPipeId)
		{
	case SMB_SRVSVC:	return &sSrvSvcItem;
		}



return 0;
}


//*****************************************************************************
//*
//*		SrvScvWrite
//*
//*****************************************************************************
//	Writes data into the SRVSVC pipe
//	pSession		: is the pointer to the smb session
//	pPipe			: is the pointer to the pipe 
//	pData			: data which should be written to the pipe
//	iLength			: is the size of the data in bytes
//	Returns the written byte or an error number below zero.
static int SrvScvWrite(SmbSession *pSession,SmbPipeEntryStr *pPipe,const void *pData,int iLength)
{
const char	   *pBuffer;		   	
DecRpcHeader   *pHeader;
SmbTrans		sTrans;
unsigned		uValue;
int				iRet;


	PRINT_SMB_MSG(("\nSRVSVC  write cmd=%i len=%i",CVAL(pData,2),iLength));


	if(iLength<20)											// check data length
		{
		PRINT_SMB_ERR(("\nSRVSVC  write wrong size (%i)",iLength));
		return SYS_ERR_PARAM;
		}

	pHeader = (DecRpcHeader*)pData;

	if(pHeader->ucVersionMajor!=5)							// check version
		{
		PRINT_SMB_ERR(("\nSRVSVC  write wrong version (%i)",pHeader->ucVersionMajor));
		return SYS_ERR_PARAM;
		}

	if((pHeader->ucPacketFlags&3)!=0x03)					// message must be a full block
		{
		PRINT_SMB_ERR(("\nSRVSVC  write not full buffer (%02Xh)",pHeader->ucPacketFlags));
		return SYS_ERR_PARAM;
		}

	if(!SmbIpcLock(pSession,IPC_BUFF_WAIT))					// lock the buffer
		{
		PRINT_SMB_ERR(("\nSRVSVC  write can't lock buffer"));
		return SYS_ERR_NO_HANDLES;
		}
	
	
	switch(pHeader->ucPacketType)
		{
	case DCERPC_PKT_BIND:									// bind

			pBuffer	= "\x05\x00\x0C\x03\x10\x00\x00\x00\x44\x00\x00\x00\x02\x00\x00\x00"
					  "\xB8\x10\xB8\x10\xF0\x53\x00\x00\x0D\x00\x5C\x50\x49\x50\x45\x5C"
					  "\x73\x72\x76\x73\x76\x63\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00"
					  "\x04\x5D\x88\x8A\xEB\x1C\xC9\x11\x9F\xE8\x08\x00\x2B\x10\x48\x60"
					  "\x02\x00\x00\x00";

			memcpy(aIpcBuffer,pBuffer,68);
			uValue = pHeader->uCallId;
			SIVAL(aIpcBuffer,12,uValue);

			uIpcBufferMode	= 2;
			uIpcBufferRd	= 0;
			uIpcBufferWr	= 68;

			iRet = iLength;
			break;

	case DCERPC_PKT_REQUEST:								// is it an srvsvc request

			memset(&sTrans,0,sizeof(sTrans));

			sTrans.iTotalDataSend	= iLength;
			sTrans.iMaxDataReturn	= sizeof(aIpcBuffer);
			sTrans.iOutputSize		= sizeof(aIpcBuffer);
			sTrans.pOutput			= aIpcBuffer;
			sTrans.pInput			= (unsigned char*)pData;

			   iRet = SmbSrvSvc(pSession,&sTrans);
			if(iRet<0 || sTrans.iDataReturn>sizeof(aIpcBuffer))
				{
				SmbFreeTrans(&sTrans);
				iRet = SYS_ERR_UNKNOWN;	
				break;
				}

			uIpcBufferMode			= 2;
			uIpcBufferRd			= 0;
			uIpcBufferWr			= sTrans.iDataReturn;

			if(sTrans.ucDataOutMode)
				{
				memcpy(aIpcBuffer,sTrans.pDataOut,uIpcBufferWr);
				SmbFreeTrans(&sTrans);
				}
			
			iRet = iLength;
			break;			 

	default:												// unknown command
			iRet = SYS_ERR_PARAM;	
		}			

	
	SysSemaphoreUnlock(hIpcSem);

	
return iRet;							
}

//*****************************************************************************
//*
//*		SrvScvRead
//*
//*****************************************************************************
//	Writes data from the SRVSVC pipe
//	pSession		: is the pointer to the smb session
//	pPipe			: is the pointer to the pipe 
//	pData			: data which should be written to the pipe
//	iLength			: is the size of the data in bytes
//	Returns the written byte or an error number below zero.
static int SrvScvRead(SmbSession *pSession,SmbPipeEntryStr *pPipe,void *pData,int iLength)
{
int				iByte;



	PRINT_SMB_MSG(("\nSRVSVC  read cmd=%i len=%i",CVAL(pData,2),iLength));

	SysSemaphoreLock(hIpcSem);
															// is the pipe ready for this request
	if(uIpcBufferMode!=2 || (uIpcBufferSession!=pSession->uId && wIpcBufferTid!=pSession->wTid && wIpcBufferFid!=pSession->wFid))
		{
		PRINT_SMB_ERR(("\nSRVSVC  wrong buffer (mode=%u,%i,%i,%i)",uIpcBufferMode,uIpcBufferSession!=pSession->uId,wIpcBufferTid!=pSession->wTid,wIpcBufferFid!=pSession->wFid));
		SysSemaphoreUnlock(hIpcSem);
		return SYS_ERR_ACCESS_DENIED;
		}

	iByte = uIpcBufferWr-uIpcBufferRd;
	
	if(iLength>iByte)										// check data size in buffer
		{
		iLength = iByte;
		}

	memcpy(pData,aIpcBuffer,iLength);

	uIpcBufferRd += iLength;

	if(uIpcBufferRd>=uIpcBufferWr)							// all data form buffer tranvered	
		{
		uIpcBufferWr		= 0;
		uIpcBufferRd		= 0;
		uIpcBufferMode		= 0;  
		uIpcBufferSession	= 0;
		SysSemaphoreUnlock(hIpcLock);
		}

	SysSemaphoreUnlock(hIpcSem);



return iLength;
}


//*****************************************************************************
//*
//*		GetStringW
//*
//*****************************************************************************
//	Gets an Unicode String
//	pData		: pointer to the data block
//	uOff		: offset of the string header
//	uStart		: offset of the string data			
//	uSize		: size of the data block
//	pText		: here the text will be stored (0=only get size)
//	iMax		: maximum text size
//	Returns the length of the buffer or an error code below zero
static int GetStringW(unsigned char *pData,unsigned uHeader,unsigned uStart,unsigned uSize,char *pText=0,int iMax=0)
{
DecStringW	   *pString;
unsigned short *pChars;
unsigned		uAll;
unsigned		uNum;
unsigned		uMax;
unsigned		uOff;
	
	
	
	if(uHeader+sizeof(DecStringW)>uSize)					// check buffer size
		{
		return WERR_INVALID_PARAM; 
		}

	pString	= (DecStringW*)(pData+uHeader);
	uAll	= ((pString->uMax+1)&~1)*2;
	uNum	= pString->uNum;
	uOff	= pString->uOff;
	uMax	= pString->uMax;
															
	if(uStart+uAll>uSize || uNum==0 || uMax<uOff+uNum)		// check sting data
		{
		return WERR_INVALID_PARAM; 
		}
	
	pChars	= (unsigned short*)(pData+uStart+uOff*2);

	if(pText)												// get text 	
		{
		#if SMB_UNICODE
		SysWideToUtf8(pText,pChars,iMax,uNum-1);			
		#else
		StrWideToChar(pText,pChars,iMax,uNum-1);
		#endif 	
		}


return uAll;
}


//*****************************************************************************
//*
//*		SetStringW
//*
//*****************************************************************************
//	Gets an Unicode String
//	pData		: pointer to the data block
//	pText		: here the text will be stored
//	Returns the length of the buffer
static int SetStringW(unsigned char *pData,const char *pText)
{
DecStringW	   *pString;
unsigned short *pChars;
unsigned		uLen;
	
	
	pString	= (DecStringW*)(pData);
	pChars	= (unsigned short*)(pString+1);

	#if SMB_UNICODE											// copy text
	uLen	= SysUtf8ToWide(pChars,pText,256)+1;
	#else
	uLen	= StrCharToWide(pChars,pText,256)+1;
	#endif 	

	pString->uMax	= uLen;
	pString->uNum	= uLen;
	pString->uOff	= 0;

	if(uLen&1)												
		{													// fill to 4 byte 
		pChars[uLen] = 0;
		uLen++;
		}
		

return sizeof(DecStringW)+uLen*2;
}

//*****************************************************************************
//*
//*		GetShare
//*
//*****************************************************************************
//	Claculates the maximum size of a and ste the schare data
//	pSession		: is the pointer to the smb session
//	iNr				: is the share number
//	iLevel			: is the info level (0,1,501)
//	pOffset			: here the last offste will be stored
//	pData			: is the target buffer for the share (NULL=only calc size)
//	pBaseId			: pointer to string ids
static int GetShare(SmbUserEntry *pUser,int iNr, int iLevel,unsigned *pOffset=0,unsigned char *pData=0,unsigned *pBaseId=0)
{
const char     *pService;
const char     *pComment;
unsigned		uAttribut;
unsigned		uOffset;
int				iStruct;
int				iLen2;
int				iLen1;


    switch(iLevel)											// info level 
        {
    case   0: iStruct =  4; break;
    case   1: iStruct = 12; break;
    case 501: iStruct = 16; break;
    default:  return -1;
        }

    if(iNr>=pUser->iCount)									// use ipc entry
        {
        pService = "IPC$";
        pComment = "Remote-IPC";
        }
    else{
        pService = pUser->pService[iNr];
        pComment = pUser->pComment[iNr];

        if(!pComment)pComment="";
        }


	if(!pData)												// only calc size
		{
		if(pOffset)*pOffset=iStruct;

		iLen1	 = strlen(pService);
		iLen1	 = ((iLen1+2)&~1)*2;
		iStruct += iLen1+sizeof(DecStringW);

		if(iLevel)											// add size off comment
			{
			iLen2	 = strlen(pComment);
			iLen2	 = ((iLen2+2)&~1)*2;
			iStruct += iLen2+sizeof(DecStringW);
			}
		
		return iStruct;
		}


	*(unsigned*)(pData)		  = *pBaseId;					// share name
	*pBaseId			     +=4;

	if(iLevel!=0)											
		{													// type
		*(unsigned*)(pData+4) = (iNr<pUser->iCount)? STYPE_DISKTREE:STYPE_IPC|STYPE_HIDDEN;		
		*(unsigned*)(pData+8) = *pBaseId;					// comment
		*pBaseId			 +=4;
		}

	if(iLevel==501)
		{
		if(iNr>=pUser->iCount)
			{
			uAttribut	= PERM_ACCESS_READ|PERM_ACCESS_WRITE|PERM_ACCESS_EXEC;		   
			}
		else if(pUser->cMode[iNr]==SMB_MODE_READWRITE)
            {
			uAttribut	= PERM_ACCESS_READ|PERM_ACCESS_WRITE;		   
            }
		else{
			uAttribut	= PERM_ACCESS_READ;		   
			}

		*(unsigned*)(pData+12) = uAttribut;					// permission  
		}


	pData	+= *pOffset;
	uOffset  =  0;										
	uOffset +=  SetStringW(pData+uOffset,pService);			// save service name
	
	if(iLevel!=0)
		{
		uOffset +=  SetStringW(pData+uOffset,pComment);		// save comment name
		}

		
			
return uOffset+iStruct;	
}

//*****************************************************************************
//*
//*		SrvSvcNetShareEnumAll
//*
//*****************************************************************************
//	NetShareEnumAll command for SRVSCT pipe.
//	pSession		: is the pointer to the smb session
//	pTrans			: a pointer to the request data
//	Returns the output size or -1 if the call ins't supported
static int SrvSvcNetShareEnumAll(SmbSession *pSession,SmbTrans *pTrans)
{
SmbUserEntry   *pUser;
unsigned char  *pData;
unsigned		uLevel;
unsigned		uStep;
unsigned		uSize;
unsigned		uBase;
unsigned		uOff;
unsigned		uPos;
int				iRet;
int				iAll;
int				iNr;



	uSize	= pTrans->iTotalDataSend;
	pData	= pTrans->pInput;
	uPos	= sizeof(DecRpcRequest)+4;
	uOff	= sizeof(DecStringW)+uPos;
															
	iRet	= GetStringW(pData,uPos,uOff,uSize);			// read UNC text
	if(iRet<0)return iRet;

	uOff   += iRet;
	uPos	= uOff;	

	if(uOff+4>uSize)
		{
		return WERR_INVALID_PARAM;  
		}

		pUser = pSession->pUser;
	if(!pUser || !pUser->iCount)							// exists share entries
		{
		return WERR_ACCESS_DENIED; 
		}


	uLevel	= *(unsigned*)(pData+uOff); 

	   iRet = GetShare(pUser,pUser->iCount,uLevel,&uStep);
	if(iRet<0)
		{
		return WERR_UNKNOWN_LEVEL;
		}

	iAll	= sizeof(DecRpcResponse);
	iAll   += iRet+64;

	for(iNr=0;iNr<pUser->iCount;iNr++)						// calc output size		
		{
		iAll += GetShare(pUser,iNr,uLevel); 	
		}
	
	pData	= SmbAllocData(pTrans,iAll);
	uPos	= sizeof(DecRpcResponse); 
	uBase	= 0x00020008;									// REM AZ should start after input id


	*(unsigned*)(pData+uPos) = uLevel;	uPos+=4;			// set header of share enum
	*(unsigned*)(pData+uPos) = 1;		uPos+=4;				
	*(unsigned*)(pData+uPos) = uBase;	uPos+=4; uBase+=4;
	*(unsigned*)(pData+uPos) = iNr+1;	uPos+=4;	
	*(unsigned*)(pData+uPos) = uBase;	uPos+=4; uBase+=4;
	*(unsigned*)(pData+uPos) = iNr+1;	uPos+=4;

	uOff = (iNr+1)*uStep;
	
	for(iNr=0;iNr<=pUser->iCount;iNr++)						// copy share info arrays
		{
		uOff += GetShare(pUser,iNr,uLevel,&uOff,pData+uPos,&uBase); 	
		uOff -= uStep*2;										
		uPos += uStep;										// jump to next array item
		}

	uPos  += uOff;	

	*(unsigned*)(pData+uPos) = iNr;		uPos+=4;			// count of shares
	*(unsigned*)(pData+uPos) = 0;		uPos+=4;			// pointer to resume handle	
	*(unsigned*)(pData+uPos) = WERR_OK;	uPos+=4;			// error code
	
	pTrans->iDataReturn	= uPos;


return WERR_OK;
}

//*****************************************************************************
//*
//*		SrvSvcNetShareGetInfo
//*
//*****************************************************************************
//	NetShareGetInfo command for SRVSCT pipe.
//	pSession		: is the pointer to the smb session
//	pTrans			: a pointer to the request data
//	Returns the output size or -1 if the call ins't supported
static int SrvSvcNetShareGetInfo(SmbSession *pSession,SmbTrans *pTrans)
{
char			cShare[64];
SmbUserEntry   *pUser;
unsigned char  *pData;
unsigned		uLevel;
unsigned		uStep;
unsigned		uSize;
unsigned		uBase;
unsigned		uOff;
unsigned		uPos;
int				iRet;
int				iAll;
int				iNr;



	uSize	= pTrans->iTotalDataSend;
	pData	= pTrans->pInput;
	uPos	= sizeof(DecRpcRequest)+4;
	uOff	= sizeof(DecStringW)+uPos;
															
	iRet	= GetStringW(pData,uPos,uOff,uSize);			// read UNC text
	if(iRet<0)return iRet;

	uOff   += iRet;
	uPos	= uOff;	
	uOff	= sizeof(DecStringW)+uPos;
															// read share name
	iRet	= GetStringW(pData,uPos,uOff,uSize,cShare,sizeof(cShare));			
	if(iRet<0)return iRet;

	uOff   += iRet;

	if(uOff+4>uSize)
		{
		return WERR_INVALID_PARAM;  
		}

		pUser = pSession->pUser;
	if(!pUser || !pUser->iCount)							// exists share entries
		{
		return WERR_ACCESS_DENIED; 
		}


	uLevel	= *(unsigned*)(pData+uOff); 


	for(iNr=0;;iNr++)										// find the share
		{
		if(iNr>=pUser->iCount)
			{
			if(!StrCmpI(cShare,"$IPC"))break;
			return WERR_BAD_NETPATH;
			}

		if(!StrCmpI(cShare,pUser->pService[iNr]))
			{
			break;
			}
		}

	   iRet = GetShare(pUser,iNr,uLevel,&uStep);
	if(iRet<0)
		{
		return WERR_UNKNOWN_LEVEL;
		}

	iAll	= sizeof(DecRpcResponse);
	iAll   += iRet+32;
	pData	= SmbAllocData(pTrans,iAll);
	uPos	= sizeof(DecRpcResponse); 
	uBase	= 0x00020004;									// REM AZ should start after input id


	*(unsigned*)(pData+uPos) = 1;		uPos+=4;				
	*(unsigned*)(pData+uPos) = uBase;	uPos+=4; uBase+=4;				

	uPos += GetShare(pUser,iNr,uLevel,&uStep,pData+uPos,&uBase); 	

	*(unsigned*)(pData+uPos) = WERR_OK;	uPos+=4;			// error code
	
	pTrans->iDataReturn	= uPos;


return WERR_OK;
}

//*****************************************************************************
//*
//*		SmbSrvSvc
//*
//*****************************************************************************
//	Will be called for trans requests on the SRVSCT pipe.
//	pSession		: is the pointer to the smb session
//	pTrans			: a pointer to the request data
//	Returns the output size or -1 if the call ins't supported
int SmbSrvSvc(SmbSession *pSession,SmbTrans *pTrans)
{
DecRpcResponse	   *pResponse;
DecRpcRequest 	   *pRequest;
DecRpcFaultResp	   *pFault;
unsigned			uErr;




	
	if(pTrans->iTotalDataSend<sizeof(DecRpcRequest))		// check header size
		{
		return -1;
		}
	
	pRequest	= (DecRpcRequest*)(pTrans->pInput);
	uErr		= WERR_OK;

	if(pRequest->uAllocHint+sizeof(DecRpcRequest)>pRequest->sHeader.usFragLength)
		{
		uErr = WERR_INVALID_PARAM;
		goto Error;
		}

	if(pRequest->sHeader.ucPacketType!=DCERPC_PKT_REQUEST)
		{
		uErr = WERR_INVALID_PARAM;
		goto Error;
		}


	switch(pRequest->usOperator)							// execute command
		{
	case SRVSVC_NETSHAREGETINFO:

			uErr = SrvSvcNetShareGetInfo(pSession,pTrans);
			break;

	case SRVSVC_NETSHAREENUMALL:
			
			uErr = SrvSvcNetShareEnumAll(pSession,pTrans);
			break;
	
	default:
			uErr = WERR_CALL_NOT_IMPLEMENTED;
			goto Error;	
		}


	if(uErr!=WERR_OK)										// send error message
		{
Error:	if(pTrans->iDataReturn==0)							// no buffer allocated
			{
			pResponse = (DecRpcResponse*)SmbAllocData(pTrans,sizeof(DecRpcResponse)+sizeof(DecRpcFaultResp));
			}
		else if(pTrans->iDataReturn>=sizeof(DecRpcResponse)+sizeof(DecRpcFaultResp))
			{												// use the existing buffer
			pTrans->iDataReturn	= sizeof(DecRpcResponse)+sizeof(DecRpcFaultResp);
			pResponse = (DecRpcResponse*)(pTrans->pParamOut);
			}
		else{												// alloc a new data
			pResponse = (DecRpcResponse*)SmbAllocData(pTrans,sizeof(DecRpcResponse)+sizeof(DecRpcFaultResp));
			if(!pTrans->ucDataOutMode)pTrans->ucDataOutMode=1;
			}

		pFault = (DecRpcFaultResp*)(pResponse+1);
		pFault->uStatus					= uErr;
		pFault->uReserved				= 0;
		pTrans->iParamReturn			= 0;
		}
	else{
		pResponse = (DecRpcResponse*)pTrans->pDataOut;
		}			

	
	memcpy(pResponse,pRequest,sizeof(DecRpcResponse));

	pResponse->sHeader.ucPacketFlags	= (uErr)? 0x23:0x03;
	pResponse->sHeader.ucPacketType		= (uErr)? DCERPC_PKT_FAULT:DCERPC_PKT_RESPONSE;
	pResponse->sHeader.usFragLength		= pTrans->iDataReturn;
	pResponse->uAllocHint				= pTrans->iDataReturn-sizeof(DecRpcResponse);
	pResponse->ucReserved				= 0;
	pResponse->ucCancelCount			= 0;


return 0;
}
