//*****************************************************************************
//*
//*
//*     SystemLinux.cpp
//*
//*
//*****************************************************************************
//
//  This file is an implementation of system functions for LINUX
//
//  Copyright  2003    Anton Zechner
//
#include <time.h>
#include <fcntl.h>
#include <sched.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <malloc.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/utime.h>
#include <sys/types.h>
#include <sys/errno.h>
#include "freertos/FreeRTOS.h"
#include "freertos/semphr.h"
#include "freertos/task.h"
#include "System/System.h"
#include "esp_vfs_fat.h"
#include "driver/sdmmc_host.h"
#include "driver/sdmmc_defs.h"
#include "sdmmc_cmd.h"


#define     HAS_PATH_CACHE          0               // cache converted paresed paths
#define     MAP_DOS_FILE_BITS       0               // use execute bits to simulate HIDDEN SYSTEM and ACHIVE flags
#define     MAX_EX_HANDELS          50              // the count of similar open handels
#define     MAX_TEMP_PATH           256             // is the maximal path size whith can be converted
#define     SYS_API                 extern "C"      // call convention of the functions

static      int                     iTimeInit=0;    // is the time zone initalized
static      int                     iHandleInit=0;  // is the handle mutex initalized
static      SemaphoreHandle_t         hHandleMutex;   // is the handle mutex for all handles

#define     HANDLE_INIT()           if(!iHandleInit){hHandleMutex=xSemaphoreCreateMutex(); iHandleInit=1;}
#define     HANDLE_LOCK()           xSemaphoreTake(hHandleMutex,(TickType_t)0);
#define     HANDLE_UNLOCK()         xSemaphoreGive(hHandleMutex);

#if         SYS_HAS_NAME_INFO

    #define EXDENDED_HANDLES        1

#endif

#if         MAP_DOS_FILE_BITS

    #define EXDENDED_HANDLES        1

#endif

#if EXDENDED_HANDLES

    typedef struct
        {        
        FIL     *iHandle;
        #if     MAP_DOS_FILE_BITS
        int     iArchBit;
        #endif
		char   *pName;
        }ExHandle;

    static ExHandle             aHandles[MAX_EX_HANDELS];
    static unsigned volatile    uNextPos=0;
#define     FIND_MAX    10                              // count of maximal similar open find handles
#define     FIND_LEN    256                             // maximal size of a find path
#define     FIND_ONE    (FIND_MAX+12345)	

    static void SysDelExtHandle(SysFile hHandle)
                {
                unsigned    uPos=(unsigned)(hHandle)-1;

                if(uPos<MAX_EX_HANDELS)
                if(aHandles[uPos].pName)
                    {
                    #if SYS_HAS_NAME_INFO
                    free(aHandles[uPos].pName);
                    #endif
		   			free(aHandles[uPos].iHandle);
                    aHandles[uPos].pName = 0;
                    }
                }


static	SysFile SysNewExtHandle(FIL iHandle,const char *pName,int *pError)
                {
				FIL		*pHandle;
                unsigned    uCount;
                unsigned    uPos;
                int         iLen;

                HANDLE_INIT();
                
				pHandle=(FIL*)malloc(sizeof(FIL));
				*pHandle=iHandle;

				#if SYS_HAS_NAME_INFO	
                
                for(uCount=0;uCount<MAX_EX_HANDELS;uCount++)
                    {
                    uPos = uNextPos%MAX_EX_HANDELS;
                    uNextPos++;

                    HANDLE_LOCK();

                    if(aHandles[uPos].pName)
                        {
                        HANDLE_UNLOCK();
                        continue;
                        }

				iLen=strlen(pName);
                aHandles[uPos].pName    = (char*)malloc(iLen);
				
                if(!aHandles[uPos].pName)
                    {
                    if(pError)*pError=SYS_ERR_UNKNOWN;
                    f_close(&iHandle);
                    return 0;
                    }
				strcpy(aHandles[uPos].pName,pName);
				#else

                aHandles[uPos].pName = "";

                #endif

				aHandles[uPos].iHandle  = pHandle;

				HANDLE_UNLOCK();
				
                return (SysFile)(uPos+1);
                }

                if(pError)*pError=SYS_ERR_NO_HANDLES;
                f_close(&iHandle);

                return 0;
                }

    #define SYS_GET_NAME(h)         (((unsigned)((unsigned)(h)-1)>=MAX_EX_HANDELS)? "":aHandles[(unsigned)(h)-1].pName)
    #define SYS_GET_HANDLE(h)       (((unsigned)((unsigned)(h)-1)>=MAX_EX_HANDELS)? 0:aHandles[(unsigned)(h)-1].iHandle)
    //#define SYS_GET_HANDLE(h)       (aHandles[(unsigned)(h)-1].iHandle)
    #define SYS_NEW_HANDLE(h,n,e)   SysNewExtHandle(h,n,e)
    #define SYS_DEL_HANDLE(h)       SysDelExtHandle(h);

    #if     MAP_DOS_FILE_BITS
    #define SYS_SET_ARCHBIT(h,s)    if(s>0 && ((unsigned)((unsigned)(h)-1)<MAX_EX_HANDELS))aHandles[(unsigned)(h)-1].iArchBit=1;
    #else
    #define SYS_SET_ARCHBIT(h,s)
    #endif

#else

    #define SYS_GET_NAME(h)         ""
    #define SYS_GET_HANDLE(h)       (((int)h)-1)
    #define SYS_NEW_HANDLE(h,n,e)   ((SysFile)(h+1))
    #define SYS_DEL_HANDLE(h)
    #define SYS_SET_ARCHBIT(h,s)

#endif
typedef struct
   {
   int                  iFree;                          // is the handle in use
   int                  iLen;                           // is the length of the find path
   int                  iNum;                           // is the length of the seach name
   FF_DIR                 *pDir;                           // is the pointer to the directory entry
   char                 cFind[FIND_LEN+4];              // is the seach name with wildchards
   char                 cName[FIND_LEN+4];              // is the path name of the search
   }FindMem;


static  FindMem         aFindMem[FIND_MAX];
static  int             iFindInit=0;
static unsigned volatile    uNextFindPos=0;

/*static SysFind SysNewFindHandle(FF_DIR *iHandle,int *pError)
                {
                unsigned    uCount;
                unsigned    uPos;
		FIND_HANDLE_INIT();
                for(uCount=0;uCount<FIND_MAX;uCount++)
                    {
                    uPos = uNextFindPos%FIND_MAX;
                    uNextFindPos++;
                    FIND_HANDLE_LOCK();

                    if(aFindMem[uPos].iFree)
                        {
                        FIND_HANDLE_UNLOCK();
                        continue;
                        }

                    aFindMem[uPos].iFree    = 1;
                    aFindMem[uPos].pDir  = iHandle;
                    FIND_HANDLE_UNLOCK();

                    return (SysFind)(uPos+1);
                    }
		if(pError)*pError=SYS_ERR_NO_HANDLES;
                f_closedir(iHandle);
                return 0;
                }
*/
#if         HAS_PATH_CACHE

    #error  "isn't implemented"

#else

    #define CALC_CHKSUM(a,b)
    #define CACHE_ADD(a,b,c)

#endif


typedef struct
    {
    int       iPriority;                                // is the thread priority (not supported)
    int       iStartOk;                                 // will be set to 1 if thread was stared
    void    (*pStartProc)(void*);                       // is a pointer to the therad function
    void     *pStartArg;                                // is the arument pointer for the thread
    }ThreadData;

typedef struct
    {
    int                 iId;                            // is the id of the semaphore 0x1255AA78
    int                 iValue;                         // is the current lock count (>0 = unlocked)
    int                 iNumber;                        // is an access count
    SemaphoreHandle_t     hAccess;                        // is the access mutex of the semathor
    SemaphoreHandle_t     hDummy;                         // is an internal used mutex
    SemaphoreHandle_t      hWait;                          // is the global wait semathor
    }SemData;





//*****************************************************************************
//*
//*     SysError
//*
//*****************************************************************************
//  Converts an Linunx ErrorCode to a System-Error-Code
static int SysError(FRESULT iOk)
{
    switch(iOk)
        {
case	FR_DISK_ERR				: return SYS_ERR_DISKFULL;			/* (1) A hard error occurred in the low level disk I/O layer */
//	FR_INT_ERR,				/* (2) Assertion failed */
//	FR_NOT_READY,			/* (3) The physical drive cannot work */
case	FR_NO_FILE				: return SYS_ERR_FILE_NOT_FOUND;/*SYS_ERR_ACCESS_DENIED;*/				/* (4) Could not find the file */
case	FR_NO_PATH				: return SYS_ERR_PATH_NOT_FOUND;			/* (5) Could not find the path */
case	FR_INVALID_NAME			: return SYS_ERR_FILE_NOT_FOUND;		/* (6) The path name format is invalid */
case	FR_DENIED				: return SYS_ERR_ACCESS_DENIED;					/* (7) Access denied due to prohibited access or directory full */
case	FR_EXIST					: return SYS_ERR_FILE_NOT_FOUND;				/* (8) Access denied due to prohibited access */
case	FR_INVALID_OBJECT		: return SYS_ERR_NO_HANDLES;	/* (9) The file/directory object is invalid */
case	FR_WRITE_PROTECTED		: return SYS_ERR_ACCESS_DENIED;		/* (10) The physical drive is write protected */
case	FR_INVALID_DRIVE			: return SYS_ERR_PARAM;		/* (11) The logical drive number is invalid */
//	FR_NOT_ENABLED,			/* (12) The volume has no work area */
//	FR_NO_FILESYSTEM,		/* (13) There is no valid FAT volume */
//	FR_MKFS_ABORTED,		/* (14) The f_mkfs() aborted due to any problem */
case	FR_TIMEOUT				: return SYS_ERR_ACCESS_DENIED;				/* (15) Could not get a grant to access the volume within defined period */
case	FR_LOCKED				: return SYS_ERR_ACCESS_DENIED;				/* (16) The operation is rejected according to the file sharing policy */
case	FR_NOT_ENOUGH_CORE		: return SYS_ERR_DISKFULL;		/* (17) LFN working buffer could not be allocated */
case	FR_TOO_MANY_OPEN_FILES	: return SYS_ERR_PARAM;	/* (18) Number of open files > _FS_LOCK */
case	FR_INVALID_PARAMETER		: return SYS_ERR_PARAM;
default:	return SYS_ERR_UNKNOWN;
        }
}

//*****************************************************************************
//*
//*      strfcmp / strnxcmp
//*
//*****************************************************************************
//  Compares to strings, the second string could contain the wildchar ?.
//  pText       : is the pointer to the first string.
//  pSource     : is the pointer to the second string.
//                a '?' represents an sigle char.
//  iLen        : is the length of the strings in byte
//  returns         0 if the strings are equal
//                  1 if the strings are not equal
static int strfcmp(const char *pText,const char *pSource,int iLen)
{
register int    i;
register char   a,b;


    for(i=0;i<iLen;i++)                         // Compare strings
        {
        b=pSource[i];
        if(b=='?')continue;
        a=pText[i];
        if(a>='a' && a<='z')a+='A'-'a';
        if(b>='a' && b<='z')b+='A'-'a';
        if(a!=b)return  1;
        }


return 0;
}

//*****************************************************************************
//*
//*      strnxcmp
//*
//*****************************************************************************
//  Compares to strings, the second string could contain the wildchars * and ?
//  pText       : is the pointer to the first string.
//  iTextLen    : is the length of the first strings in byte
//  pSource     : is the pointer to the second string.
//                a '?' represents an sigle char.
//                a '*' represents several chars.
//  iLen        : is the length of the second strings in byte
//  returns         0 if the strings are equal
//                  1 if the strings are not equal
static int strnxcmp(const char *pText,int iTextLen,const char *pSource,int iSourceLen)
{
//int     iTextPos,iFindPos;
int     i,iLen,iStar;



    if(iSourceLen<=0)return 0;


//    iFindPos=0;
//    iTextPos=0;


    for(;;)                                     // Suchen Schleife
        {
        if(iSourceLen<=0)break;
        if(  iTextLen<=0)break;

        iLen=0;
        while(iLen<iSourceLen)                  // Wort Lnge holen
            {
            if(pSource[iLen]=='*')break;
            iLen++;
            }
        if(iTextLen<iLen)return 1;
        if(iLen)
            {
            if(strfcmp(pText,pSource,iLen))return 1;
            iSourceLen-=iLen;pSource+=iLen;
              iTextLen-=iLen;  pText+=iLen;
            continue;
            }

        iStar=0;

        while(iSourceLen>0)                     // Wildcard berspringen
            {
            if(*pSource=='?'){iStar++;pSource++;iSourceLen--;continue;}
            if(*pSource!='*')break;
                pSource++;iSourceLen--;
            }

        if(iTextLen<iStar)return 1;
        if(iSourceLen<=0 )return 0;

        iTextLen -= iStar;
        pText    += iStar;
        iLen      = 0;

        while(iLen<iSourceLen)                  // Wort Lnge holen
            {
            if(pSource[iLen]=='*')break;
            iLen++;
            }

        while(iTextLen>=iLen)                   // Source Wortteil
            {
            if(!strfcmp(pText,pSource,iLen))
                {
                if(iSourceLen==iLen)return iTextLen!=iLen;
                i=strnxcmp( pText  +iLen,  iTextLen -iLen,
                            pSource+iLen,iSourceLen -iLen);

                if(i==0)return i;
                }

            iTextLen--;
            pText++;
            }


        return 1;
        }

//********************** Suchen_len <= 0 **************************************

    if(iSourceLen>0)                                // Stern am Wortende
        {
        while(iSourceLen>0)
            {
            if(*pSource!='*')return 1;
            iSourceLen--;
            pSource--;
            }
        return 0;
        }

    if(iTextLen>0)return 1;

return 0;
}
#if SYS_USE_UTF8NAMES

//*****************************************************************************
//*
//*     SysUtf8ToChar
//*
//*****************************************************************************
//  Wandelt das erste UTF8-Zeichen aus einem Text in ein Zeichen um
//  pText       : Zeiger auf UTF8 Text
//  uUnicode    : Hier wird das Zeichen gespeichert
//  Ergibt die Lnge des Zeichens
static inline int SysUtf8ToChar(const char *pText,unsigned &uUnicode)
{
char        cSign1;
char        cSign2;
char        cSign3;
char        cChar;


    cChar=*pText;

    if((cChar&0xC0)==0xC0)
        {
        if((cChar&0xE0)==0xC0)                  // Zwei Zeichen ?
            {
            cSign1 = pText[1];

            if((cSign1&0xC0)==0x80)
                {
                uUnicode = ((cChar&0x1F)<<6)|(cSign1&0x3F);
                return 2;
                }
            }
        else if((cChar&0xF0)==0xE0)             // Drei Zeichen ?
            {
            cSign1 = pText[1];
            cSign2 = pText[2];

            if((cSign1&0xC0)==0x80)
            if((cSign2&0xC0)==0x80)
                {
                uUnicode = ((cChar&0x0F)<<12)|((cSign1&0x3F)<<6)|(cSign2&0x3F);
                return 3;
                }
            }
        else if((cChar&0xF8)==0xF0)             // Vier Zeichen ?
            {
            cSign1 = pText[1];
            cSign2 = pText[2];
            cSign3 = pText[3];

            if((cSign1&0xC0)==0x80)
            if((cSign2&0xC0)==0x80)
            if((cSign3&0xC0)==0x80)
                {
                uUnicode = ((cChar&0x07)<<18)|((cSign1&0x3F)<<12)|((cSign2&0x3F)<<6)|(cSign3&0x3F);
                return 4;
                }
            }
        }


    uUnicode = (unsigned char)cChar;


return 1;
}

//*****************************************************************************
//*
//*     SysCharToWide
//*
//*****************************************************************************
//  Wandelt ein Zeichen in einen UTF16 Text um
//  pText       : Zeiger auf UTF8/UTF16 Text wo das Zeichen gespeichert  wird
//  uUnicode    : Ist das Zeichen das umgewandelt werden soll
//  Ergibt die Lnge des Zeichens
static inline int SysCharToWide(unsigned short *pText,unsigned uUnicode)
{

    if(uUnicode<0x10000)                                    // Ein Zeichen
        {
        pText[0]=(unsigned short)uUnicode;
        return 1;
        }


    uUnicode -= 0x10000;

    pText[0] = (unsigned short)(0xD800|(uUnicode>>10   ));
    pText[1] = (unsigned short)(0xDC00|(uUnicode&0x03FF));


return 2;
}

//*****************************************************************************
//*
//*     SysCharToUtf8
//*
//*****************************************************************************
//  Wandelt ein Zeichen in eine UTF8/UTF16 Text um
//  pText       : Zeiger auf UTF8/UTF16 Text wo das Zeichen gespeichert  wird
//  uUnicode    : Ist das Zeichen das umgewandelt werden soll
//  Ergibt die Lnge des Zeichens
static inline int SysCharToUtf8(char *pText,unsigned uUnicode)
{

    if(uUnicode<0x80)                                       // Ein Zeichen
        {
        pText[0]=(char)uUnicode;
        return 1;
        }

    if(uUnicode<0x800)                                      // Zwei Zeichen
        {
        pText[0]=(char)((uUnicode>>  6)|0xC0);
        pText[1]=(char)((uUnicode&0x3F)|0x80);
        return 2;
        }

    if(uUnicode<0x10000)                                    // Drei Zeichen
        {
        pText[0]=(char)( (uUnicode>>12)      |0xE0);
        pText[1]=(char)(((uUnicode>> 6)&0x3F)|0x80);
        pText[2]=(char)( (uUnicode     &0x3F)|0x80);
        return 3;
        }

    pText[0]=(char)(((uUnicode>>18)&0x07)|0xF0);
    pText[1]=(char)(((uUnicode>>12)&0x3F)|0x80);
    pText[2]=(char)(((uUnicode>> 6)&0x3F)|0x80);
    pText[3]=(char)( (uUnicode     &0x3F)|0x80);


return 4;
}

//*****************************************************************************
//*
//*     SysUtf8ToWide
//*
//*****************************************************************************
//  Konvertiert einen UFF8 Text in einen UTF16 Text
//  pTextPtr    : Zeiger auf den Textpuffer fr den UTF16 Text.
//  pUtf8       : Zeiger auf den UTF8-String.
//  iLen        : Gre des Zielpuffers in Zeichen.
//  iTextLen    : Maximale Gre des UTF8-Strings in Byte.
//  Ergibt die Gre des kopierten Textes in Zeichen.
extern "C" int SysUtf8ToWide(void *pTextPtr,const char *pUtf8,int iLen,int iTextLen)
{
unsigned short     *pText=(unsigned short*)pTextPtr;
unsigned    uUnicode;
char        cTemp[4];
int         iOut;
int         iPos;


    iLen -= 2;

    if(iLen<0)                                              // Ist der Zielpuffer zu klein
        {
        if(iLen==-1)*pText=0;
        return 0;
        }

    for(iOut=0,iPos=0;;)
        {
        if(iOut>=iLen)goto Exit;
        if(iPos+4>iTextLen)break;
        iPos += SysUtf8ToChar(pUtf8+iPos,uUnicode);
        if(uUnicode==0)goto Exit;
        iOut += SysCharToWide(pText+iOut,uUnicode);
        }

    memcpy(cTemp,pUtf8+iPos,iTextLen-iPos);
    cTemp[iTextLen-iPos]=0;

    for(iPos=0;;)                                           // Die letzen 3 Byte kopieren
        {
        if(iOut>=iLen)break;
        iPos += SysUtf8ToChar(cTemp+iPos,uUnicode);
        if(uUnicode==0)break;
        iOut += SysCharToWide(pText+iOut,uUnicode);
        }

Exit:

    pText[iOut] = 0;



return iOut;
}


//*****************************************************************************
//*
//*     SysWideToUtf8
//*
//*****************************************************************************
//  Kopiert einen Text in einen UTF8-String.
//  pUtf8       : Zeiger auf den Textpuffer fr den UTF8-String.
//  pTextPtr    : Zeiger auf den Quelltext im UTF16 Format.
//  iLen        : Gre des Zielpuffers in Byte.
//  iTextLen    : Maximale Gre des Quelltextes in Zeichen.
//  Ergibt die Gre des UTF8-String in Byte.
extern "C" int SysWideToUtf8(char *pUtf8,const void *pTextPtr,int iLen,int iTextLen)
{
unsigned short *pText=(unsigned short*)pTextPtr;
unsigned		uUnicode;
int				iOut;
int				iPos;


    iLen -= 4;

    if(iLen<0)return 0;

    for(iOut=0,iPos=0;iOut<iLen;iPos++)
        {
        if(iPos>=iTextLen)break;
        uUnicode = pText[iPos];
        if(uUnicode==0)break;

        if((uUnicode   >>10)==(0xD800>>10) && iPos+1<iLen)
        if((pText[iPos]>>10)==(0xDC00>>10))
            {
            uUnicode   = (0x03FF&uUnicode   )<<10;
            uUnicode  |= (0x03FF&pText[iPos]);
            }

        iOut += SysCharToUtf8(pUtf8+iOut,uUnicode);
        }

    pUtf8[iOut] = 0;




return iOut;
}

#endif
//*****************************************************************************
//*
//*     SysFullPath
//*
//*****************************************************************************
//  converts a relative path to a full path
//  pPath       : is the relative path
//  pBuffer     : is the buffer for the full path
//  iMax        : is the size of the buffer
//  returns the lenght in bytes of the full path cName.
static int SysFullPath(const char *pPath,char *pBuffer,int iMax)
{
int      iLen,iSlash;
int      iPos=0,iAdd;
FRESULT iOk;


    if(iMax<=0)return 0;

    pBuffer[0]=0;
    iLen=strlen(pPath);
    if(iLen+1>=iMax)return 0;

    if(pPath[0]!='/' && pPath[0]!='\\')
        {
        iOk=f_getcwd((TCHAR*)pBuffer,iMax);
        if(iOk!=FR_OK)
	{
		return 0;
	}
        iPos=strlen(pBuffer);
        if(iPos+1+iLen>=iMax)return 0;
        pBuffer[iPos]='/';
        iPos++;
        memcpy(pBuffer+iPos,pPath,iLen+1);
        iLen+=iPos;
        }
    else{
        memcpy(pBuffer,pPath,iLen+1);
        pBuffer[0]='/';
        iPos=1;
        }


    while(iPos<iLen)
        {                                               // filter "//"
        if(iPos>0 && (pBuffer[iPos]=='/' || pBuffer[iPos]=='\\'))
            {
            memmove(pBuffer+iPos,pBuffer+iPos+1,iLen-iPos);
            iLen-=1;
            continue;
            }
                                                        // filter "./"
        if(iPos>0 && pBuffer[iPos]=='.' && (pBuffer[iPos+1]=='/' || pBuffer[iPos+1]=='\\'))
            {
            memmove(pBuffer+iPos,pBuffer+iPos+2,iLen-iPos-1);
            iLen-=2;
            continue;
            }
                                                        // filter "."
        if(iPos>1 && pBuffer[iPos]=='.' && pBuffer[iPos+1]==0)
            {
            iPos--;
            pBuffer[iPos]=0;
            iLen=iPos;
            break;
            }
                                                        // filter "../"
        if(iPos>0 && pBuffer[iPos]=='.' && pBuffer[iPos+1]=='.' && (pBuffer[iPos+2]=='/' || pBuffer[iPos+2]=='\\'))
            {
            iSlash=iPos;

            for(iPos--;iPos>0;iPos--)
                {
                if(pBuffer[iPos-1]=='/')break;
                }

            iAdd = iSlash-iPos+3;

            memmove(pBuffer+iPos,pBuffer+iSlash+3,iLen-iSlash-2);
            iLen-=iAdd;
            continue;
            }
                                                        // filter ".."
        if(iPos>1 && pBuffer[iPos]=='.' && pBuffer[iPos+1]=='.' && pBuffer[iPos+2]==0)
            {
            for(iPos-=2;iPos>0;iPos--)
                {
                if(pBuffer[iPos]=='/')break;
                }

            if(iPos<=0)
                {
                pBuffer[0]='/';
                iPos=1;
                }

            pBuffer[iPos]=0;
            iLen=iPos;
            break;
            }

        for(;pBuffer[iPos];iPos++)                      // to next slash
            {
            if(pBuffer[iPos]!='/')
                {
                if(pBuffer[iPos]!='\\')continue;
                pBuffer[iPos]='/';
                }

            iPos++;
            break;
            }
        }



return iLen;
}

//*****************************************************************************
//*
//*     SysCheckPath
//*
//*****************************************************************************
//  converts a windows path to an linux path
//  pPath   : is the pointer to the windows path
//  pBuffer : is a buffer for the new linux path
//  returns a pointer to the new path
static const char *SysCheckPath(const char *pPath,char *pBuffer)
{
	SysFullPath(pPath,pBuffer,MAX_TEMP_PATH);
	
/*	iLen=strlen(pPath);
	memcpy(pBuffer,pPath,iLen+1);
	for(i=iLen;i>0;)
	{
		i--;
		if(pBuffer[i]=='\\') pBuffer[i]='/';
	}
	pBuffer[iLen]=0;
*/
return pBuffer;	
/*
#if HAS_PATH_CACHE
unsigned        uChkSum=0;
#endif
#if SYS_HAS_SHORT_NAME
int             iNum,iPoint,iFoundLen=0;
char            cFound[sizeof(struct dirent)];
unsigned        uChkShr;
unsigned        uChkNow;
int             iShort;
#endif
FF_DIR            pDir;
FILINFO   sEntry;
FILINFO     sStat;
char            cLast;
char            cChar;
int             iSize;
int             iLen;
FRESULT            iOk;
int             i,j;
   
        iLen = SysFullPath(pPath,pBuffer,MAX_TEMP_PATH);
       iOk = f_stat(pBuffer,&sStat);
    if(iOk==FR_OK)                                          // exists the path
        {
        return pBuffer;
        }


    for(i=0;i<iLen;i=iSize)
        {
        for(iSize=i+1;iSize<iLen;iSize++)
            {
               cChar = pBuffer[iSize];
            if(cChar=='/')break;

            CALC_CHKSUM(uChkSum,cChar);
            }


        cLast           = pBuffer[iSize];
        pBuffer[iSize]  = 0;

		
           iOk = f_stat(pBuffer,&sStat);  
        if(iOk==FR_OK)                                      // exists the path
            {
            CACHE_ADD  (uChkSum,pBuffer,iSize);
            CALC_CHKSUM(uChkSum,'/'+i);
            pBuffer[iSize] = cLast;
            continue;
            }

        #if HAS_PATH_CACHE
        if(CACHE_FIND(uChkSum,pBuffer,iSize))           // is path in cache
            {
            CALC_CHKSUM(uChkSum,'/'+i);
            pBuffer[iSize] = cLast;
            continue;
            }
        #endif


        j =(!i)? 1:i;
        cChar       = pBuffer[j];
        pBuffer[j]  = 0;
        iOk        = f_opendir(&pDir,pBuffer);                 // open directory entries
        pBuffer[j]  = cChar;

        if(iOk!=FR_OK)	//cannot open dir
            {
            pBuffer[iSize] = cLast;
            return pBuffer;
            }

        i++;

        #if SYS_HAS_SHORT_NAME
           iNum = iSize-i;	//object name need to verify.
        if(iNum<=12 && iNum>=8 && pBuffer[i+2]=='~' &&
           iNum!= 9 &&(iNum==8 || pBuffer[i+8]=='.'))
            {
            uChkShr = 0;

            for(j=3;j<8;j++)
                {
                cChar = pBuffer[i+j];

                if(cChar<'0')break;
                if(cChar>'9' && cChar<'A')break;
                if(cChar>'Z' && cChar<'a')break;
                if(cChar>='a')
                    {
                    if(cChar>'z')break;
                    cChar -= 'a'-'A';
                    }

                if(cChar>='A')
                        cChar -= 'A'-10;
                else    cChar -= '0';

                uChkShr *= 36;
                uChkShr += (unsigned char)cChar;
                }


            if(j!=8 || (iNum> 9 && pBuffer[i+ 9]==' ') ||
                       (iNum>10 && pBuffer[i+10]==' ') ||
                       (iNum>11 && pBuffer[i+11]==' '))
                    iShort = 0;
            else    iShort = 1;

            cFound[0] = 0;
            }
        else{
            iShort  = 0;
            uChkShr = 0;
            }
        #endif


        for(;;)
            {
            iOk=f_readdir(&pDir,&sEntry);           // get next directory entry
            if((iOk!=FR_OK)||(sEntry.fname[0]==0))break;

            #if SYS_HAS_SHORT_NAME
            if(!strncasecmp(sEntry.fname,pBuffer+i,iNum))
                {
                if(sEntry.fname[iNum]==0)break;

                if(iNum<=9)                             // is a '.' at the end
                if(sEntry.fname[iNum  ]=='.')
                if(sEntry.fname[iNum+1]== 0 )
                    {
                    memmove(pBuffer+iSize+1,pBuffer+iSize,iLen-iSize+1);
                    iSize++;
                    iLen++;
                    break;
                    }
                }

            #else
            if(!strcasecmp(sEntry.fname,pBuffer+i))break;
            #endif

            #if SYS_HAS_SHORT_NAME
            if(!iShort || cFound[0])continue;           // searchfor a short path


                cChar = sEntry.fname[0];
            if((cChar>='a' && cChar<='z') ||
               (cChar>='A' && cChar<='Z'))
                {
                if((cChar^pBuffer[i])&~0x20)continue;
                }
            else{
                if(cChar==' ')cChar='_';
                if(cChar!=pBuffer[i])continue;
                }

                cChar = sEntry.fname[1];
            if((cChar>='a' && cChar<='z') ||
               (cChar>='A' && cChar<='Z'))
                {
                if((cChar^pBuffer[i+1])&~0x20)continue;
                }
            else{
                if(cChar==' ')cChar='_';
                if(cChar!=pBuffer[i+1])continue;
                }

            uChkNow = 0;
            iPoint  = 0;

            for(j=0;j<sEntry.fname[j];j++)             // calculate checksum
                {
                cChar = sEntry.fname[j];

                if(cChar>='a' && cChar<='z')cChar-='a'-'A';
                else if(cChar=='.')iPoint=j;

                uChkNow  = (uChkNow<<5)|(uChkNow>>27);
                uChkNow += (unsigned char)cChar;
                }

            uChkNow &= 0x01FFFFFF;

            if(uChkNow!=uChkShr)continue;

            if(iPoint+1==j)                             // last chat is a '.'
                {
                if(iNum!=8)continue;
                memcpy(cFound,sEntry.fname,j);
                iFoundLen = j;
                continue;
                }

            if(iPoint==0)                               // no suffix
                {
                if(iNum!=8)continue;
                memcpy(cFound,sEntry.fname,j);
                iFoundLen = j;
                continue;
                }

                cChar = sEntry.fname[iPoint+1];
            if((cChar>='a' && cChar<='z') ||
               (cChar>='A' && cChar<='Z'))
                {
                if((cChar^pBuffer[i+9])&~0x20)break;
                }
            else{
                if(cChar==' ')cChar='_';
                if(cChar!=pBuffer[i+9])break;
                }

            if(iPoint+2==j)                             // one byte suffix
                {
                memcpy(cFound,sEntry.fname,j);
                iFoundLen = j;
                continue;
                }

                cChar = sEntry.fname[iPoint+2];
            if((cChar>='a' && cChar<='z') ||
               (cChar>='A' && cChar<='Z'))
                {
                if((cChar^pBuffer[i+10])&~0x20)break;
                }
            else{
                if(cChar==' ')cChar='_';
                if(cChar!=pBuffer[i+10])break;
                }

            if(iPoint+3==j)                             // tow byte suffix
                {
                memcpy(cFound,sEntry.fname,j);
                iFoundLen = j;
                continue;
                }

                cChar = sEntry.fname[iPoint+2];
            if((cChar>='a' && cChar<='z') ||
               (cChar>='A' && cChar<='Z'))
                {
                if((cChar^pBuffer[i+10])&~0x20)break;
                }
            else if(iPoint+4==j)
                {
                if(cChar==' ')cChar='_';
                if(cChar!=pBuffer[i+11])break;
                }
            else if(pBuffer[i+11]!='~')
                {
                break;
                }

            memcpy(cFound,sEntry.fname,j);
            iFoundLen = j;
            #endif
            }

        f_closedir(&pDir);

        if(iOk==FR_OK)                                     // found a same uncase sensidife name
            {
            memcpy(pBuffer+i,sEntry.fname,iSize-i);
            CALC_CHKSUM(uChkSum,'/'+i);
            pBuffer[iSize] = cLast;
            continue;
            }

        #if SYS_HAS_SHORT_NAME
        if(iShort && cFound[0])                         // was a short path found
            {
            pBuffer[iSize] = cLast;
            CALC_CHKSUM(uChkSum,'/'+i);

            j = iFoundLen-(iSize-i);

            if(iLen+j<MAX_TEMP_PATH)
                {
                memmove(pBuffer+iSize+j,pBuffer+iSize,iLen-iSize+1);
                memcpy (pBuffer+i,cFound,iFoundLen);
                iSize += j;
                iLen  += j;
                continue;
                }
            }
        #endif

        pBuffer[iSize] = cLast;
        break;
        }
return pBuffer;
*/
}

//*****************************************************************************
//*
//*     SysClearFindStruct
//*
//*****************************************************************************
//  Lscht eine Find-Struktur
//  pFind   : Zeiger auf die Find-Struktur
inline void SysClearFindStruct(SysFindStruct *pFind)
{

    pFind->dwSize=0;
    pFind->cName[0]=0;
    #if SYS_HAS_SHORT_NAME
    pFind->cShortName[0]=0;
    #endif
}

//*****************************************************************************
//*
//*     SysCopyShortName
//*
//*****************************************************************************
//  Converts a filename in a MSDOS 8.3 format
#if SYS_HAS_SHORT_NAME
inline void SysCopyShortName(char *pName,const char *pFullName)
{
unsigned    uChkNow;
char        cChar;
int         iPoint;
int         iNum;
int         i;



    if(pFullName[0]== 0)                            // don't convert emty names
        {
        pName[0] = 0;
        return;
        }

    if(pFullName[0]=='.')                           // don't convert "." and ".."
        {
        if(pFullName[1]==0)
            {
            pName[0] = '.';
            pName[1] = 0;
            return;
            }

        if(pFullName[1]=='.' && pFullName[2]==0)
            {
            pName[0] = '.';
            pName[1] = '.';
            pName[2] = 0;
            return;
            }
        }


    uChkNow =  0;
    iPoint  = -1;
    iNum    =  0;

    for(i=0;pFullName[i];i++)                       // calculate checksum
        {
        cChar = pFullName[i];

             if(cChar==' '){iNum=2;}
        else if(cChar=='.'){iNum++;iPoint=i;}
        else if(cChar>='a' && cChar<='z')cChar-='a'-'A';

        if(i<12)pName[i]=cChar;

        uChkNow  = (uChkNow<<5)|(uChkNow>>27);
        uChkNow += (unsigned char)cChar;
        }


    if(iNum<2 && i<12)                              // no checksum needed
    if((iPoint<0 && i<=8) || i-iPoint<=4)
        {
        if(iPoint+1==i)i--;
        pName[i] = 0;
        return;
        }

    uChkNow &= 0x01FFFFFF;

    if(pName[0]==' ')pName[0]='_';
    if(pName[1]==' ')pName[1]='_';
                     pName[2]='~';

    for(i=7;i>=3;i--)                               // write checksum to name
        {
           cChar = (char)(uChkNow%36);
        if(cChar>=10)
                cChar+='A'-10;
        else    cChar+='0';

        pName[i]=cChar;
        uChkNow/=36;
        }

    if(iPoint<0)                                    // exists a suffix
        {
        pName[8]=0;
        return;
        }


    pName[8]='.';

    for(iPoint++,i=9;i<12;i++)                      // copy suffix
        {
                cChar=pFullName[iPoint];
             if(cChar==0)break;
        else if(cChar>='a' && cChar<='z')cChar-='a'-'A';
        else if(cChar==' ')cChar='_';

        pName[i]=cChar;
        iPoint++;
        }

    if(i==9)                                        // remove the last point
        {
        i=8;
        }
    else if(i==12 && pFullName[iPoint])             // suffix larger than 3 chars ?
        {
        pName[11]='~';
        }

    pName[i] = 0;

}
#endif

//*****************************************************************************
//*
//*     SysTimeConvert     DOSTIME --> SysUtc
//*
//*****************************************************************************
//  Konvertiert eine DOS Zeit in die Systemzeit
//  uDosTime    : Ist die DOS Zeit
//  pSysTime    : Zeiger auf Puffer fr die Systemzeit
//  Ergibt 1 wenn erfolgreich ansonsten 0
SYS_API int SysTimeConvertDS(SysDW uDosTime ,SysUtc *pSysTime)
{
static const int    aTpPDays [] = {  0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366, 366, 366,   0};
static const int    aTpDays  [] = {  0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365, 365, 365,   0};
unsigned            uVal,uYear,uDays;
unsigned short      wDate;



    wDate=(unsigned short)(uDosTime>>16);

       uYear = (wDate>>9)+1980;
    if(uYear>2038)
        {
        *pSysTime=0;
        return 0;
        }

    if(uYear&3)uDays  = aTpDays  [((wDate>>5)-1)&0xF];
    else       uDays  = aTpPDays [((wDate>>5)-1)&0xF];
               uDays += ((wDate)&0x1F)-1;
               uDays += (uYear-1970  )*365;
               uDays += (uYear-1970+1)>>2;

               uVal   = uDays*24;
    *pSysTime =uVal;


return 1;
}

//*****************************************************************************
//*
//*     SysTimeConvert     SysUtc  --> DOSTIME
//*
//*****************************************************************************
//  Konvertiert eine eine Systemzeit in eine DOS-Zeit
//  uSysTime    : Ist die Systemzeit
//  pDosTime    : Zeiger auf Puffer fr die DOS Zeit
//  Ergibt 1 wenn erfolgreich ansonsten 0
SYS_API int SysTimeConvertSD(SysUtc  uSysTime,SysDW *pDosTime)
{
struct tm      *pGlobal;
struct tm       sGlobal;
time_t          iTime;
unsigned short  wTime,wDate;


    iTime = uSysTime;

       pGlobal=gmtime_r(&iTime,&sGlobal);
    if(pGlobal==NULL || pGlobal->tm_year<80 || pGlobal->tm_year>143)
        {
        *pDosTime = 1<<5;
        return 0;
        }

    wDate = (pGlobal->tm_mday  ) | ((pGlobal->tm_mon+1)<<5) | ((pGlobal->tm_year-80)<< 9);
    wTime = (pGlobal->tm_sec>>1) | ( pGlobal->tm_min   <<5) | ( pGlobal->tm_hour    <<11);

    *pDosTime = wTime|(wDate<<16);


return 1;
}

//*****************************************************************************
//*
//*     SysTimeConvertSF        SysUtc  --> LONGTIME
//*
//*****************************************************************************
//  Konvertiert eine Systemzeit in eine Windows Filetime (100ns)
//  uSysTime    : Ist die SystemZeit
//  pWftTime    : Zeiger auf Puffer fr die Windows Filetime
//  Ergibt 1 wenn erfolgreich ansonsten 0
SYS_API int SysTimeConvertSF(SysUtc  uSysTime,void *pWftTime)
{
union   {
        long long   i64;
        unsigned    dw[2];
        }uVal;


    if(uSysTime==0xFFFFFFFF || uSysTime==0)
        {
        ((unsigned*)pWftTime)[1] = 0xFFFFFFFF;
        ((unsigned*)pWftTime)[0] = 0xFFFFFFFF;
        return 0;
        }


    uVal.i64  = uSysTime;
    uVal.i64 *= 10000000;                       // Sekunden zu 100ns umwandeln
    uVal.i64 += 0x019db1ded53e8000LL;

    ((SysDW*)pWftTime)[1] = uVal.dw[1];         // Zeit speichern
    ((SysDW*)pWftTime)[0] = uVal.dw[0];


return 1;
}


//*****************************************************************************
//*
//*     SysTimeConvertFS        LONGTIME --> SysUtc
//*
//*****************************************************************************
//  Konvertiert eine Windows Filetime in eine Systemzeit
//  uDosTime    : Ist der Zeiger auf die Windows Filetime
//  pSysTime    : Zeiger auf Puffer fr die Systemzeit
//  Ergibt 1 wenn erfolgreich ansonsten 0
SYS_API int SysTimeConvertFS(const void *pWftTime,SysDW *pSysTime)
{
union   {
        long long   i64;
        unsigned    dw[2];
        }uVal;


    uVal.i64  = *(long long*)pWftTime;
    uVal.i64 -= 0x019db1ded53e8000LL;

    if(uVal.dw[1]&0x80000000)                   // Zeit vor Mi 1.Jan.1970
        {
        *pSysTime=0;
        return 0;
        }

    uVal.i64 /= 10000000;                       // 100ns zu Sekunden umwandeln

    if(uVal.dw[1]&0x80000000)                   // ber Jahr 2106
        {
        *pSysTime=0xFFFFFFFF;
        return 0;
        }

    *pSysTime = uVal.dw[0];                     // Systemzeit speichern


return 1;
}

//******************************************************************************
//*
//*     SysTimeConvertSU        SysUtc  --> time_t
//*
//******************************************************************************
//  Konvertiert eine Systemzeit in eine UTC Zeit(time_t)
//  uSysTime    : Ist der Zeiger auf die Windows Filetime
//  pUtcTime    : Zeiger auf Puffer fr die UTC-Zeit
//  Ergibt 1 wenn erfolgreich ansonsten 0
SYS_API int SysTimeConvertSU(SysUtc  uSysTime,void *pUtcTime)
{

    *(time_t*)pUtcTime = uSysTime;


return (uSysTime==0xFFFFFFFF)? 0:1;
}

//******************************************************************************
//*
//*     SysTimeConvertUS    time_t --> SysUtc
//*
//******************************************************************************
//  Konvertiert eine UTC Zeit(time_t) in eine Systemzeit
//  pUtcTime    : Ist der Zeiger auf die UTZ-Zeit
//  pSysTime    : Zeiger auf Puffer fr die Systemzeit
//  Ergibt 1 wenn erfolgreich ansonsten 0
SYS_API int SysTimeConvertUS(const void *pUtcTime,SysUtc *pSysTime)
{

    *(SysUtc *)pSysTime = *(SysUtc *)pUtcTime;


return (*pSysTime==0xFFFFFFFF || *pSysTime==0)? 0:1;
}



//******************************************************************************
//*
//*     SysTimeGet
//*
//******************************************************************************
//  Returns the current system time
SYS_API SysUtc  SysTimeGet()
{
return time(0);
}

//*****************************************************************************
//*
//*     SysTimeToStruct
//*
//*****************************************************************************
//  Converts a SysUtc  value in a time structure
//  uTime   : global time in seconds from 1.Jan.2000
//  pData   : pointer to the time structure for the global time
//  uTime   : global time value that will converted
//  uMode   : is the conversion mode
//              0 = global time
//              1 = local time
//              2 = local time wihout daylight saving
SYS_API void SysTimeToStruct(SysTimeStruct *pData,SysUtc  uTime,unsigned uMode)
{
time_t      iUtc;
struct tm  *pTimeTm;


    if(!iTimeInit)
        {
        iTimeInit=1;
        time(0);
        }

    iUtc = uTime;

    switch(uMode)
        {
    case 2:     iUtc    -= _timezone;
    case 0:     pTimeTm  =  gmtime   (&iUtc);   break;
    case 1:     pTimeTm  =  localtime(&iUtc);   break;
    default:    memset(pData,0,sizeof(SysTimeStruct));
                return;
        }


    pData->usYear   = (unsigned short) (pTimeTm->tm_year+1900);
    pData->ucMonth  = (unsigned char ) (pTimeTm->tm_mon+1);
    pData->ucDay    = (unsigned char )  pTimeTm->tm_mday;
    pData->ucHour   = (unsigned char )  pTimeTm->tm_hour;
    pData->ucMinute = (unsigned char )  pTimeTm->tm_min;
    pData->ucSecond = (unsigned char )  pTimeTm->tm_sec;
    pData->ucIsDst  = (unsigned char )((pTimeTm->tm_isdst)? 60:0);
}

//*****************************************************************************
//*
//*     SysTimeFromData
//*
//*****************************************************************************
//  Converts the local time from a time structure in a SysUtc  value
//  pData   : pointer to the time structure
//  uMode   : is the conversion mode
//              0 = pData is a global time
//              1 = pData is a local time ucIsDst is valid
//              2 = pData is a local time ucIsDst is in valid
//  Returns the converted time or 0xFFFFFFFF at an error
SYS_API SysUtc  SysTimeFromStruct(const SysTimeStruct *pData,unsigned uMode)
{
time_t          iUtc;
struct tm       sTimeTm;


    if(!iTimeInit)
        {
        iTimeInit=1;
        time(0);
        }


    sTimeTm.tm_wday  =  0;
    sTimeTm.tm_yday  =  0;
    sTimeTm.tm_year  =  pData->usYear-1900;
    sTimeTm.tm_mon   =  pData->ucMonth-1;
    sTimeTm.tm_mday  =  pData->ucDay;
    sTimeTm.tm_hour  =  pData->ucHour;
    sTimeTm.tm_min   =  pData->ucMinute;
    sTimeTm.tm_sec   =  pData->ucSecond;

    switch(uMode)
        {
    case 0:     sTimeTm.tm_isdst =  0;
                iUtc  = mktime(&sTimeTm);
                if(iUtc<=0)return 0xFFFFFFFF;
                iUtc += _timezone;
                break;

    case 1:     sTimeTm.tm_isdst =  (pData->ucIsDst)? 60:0;
                iUtc = mktime(&sTimeTm);
                break;

    case 2:     sTimeTm.tm_isdst = -1;
                iUtc = mktime(&sTimeTm);
                break;

    default:    return 0xFFFFFFFF;
        }

    if(iUtc<=0)return 0xFFFFFFFF;


return iUtc;
}

//*****************************************************************************
//*
//*     SysTimeToWeekday
//*
//*****************************************************************************
//  Converts a systime to the weekday
//  uSysTime    : is the systime value
//  Returns the weekday of the systemtime (0..6, sunday=0)
SYS_API SysDW SysTimeToWeekday(SysUtc  uSysTime)
{
unsigned    uTime;

    if(!iTimeInit)
        {
        iTimeInit=1;
        time(0);
        }
    uTime = uSysTime-_timezone;
return ((uTime/(3600*24)+4))%7;
}

//*****************************************************************************
//*
//*     SysGetFullPath
//*
//*****************************************************************************
//  converts a relative path to a full path
//  pPath       : is the relative path
//  pBuffer     : is the buffer for the full path
//  iMax        : is the size of the buffer
//  returns the lenght in bytes of the full path cName.
int SysGetFullPath(const char *pPath,char *pBuffer,int iMax)
{
int      iLen,iSlash;
int      iPos=0,iAdd;
FRESULT iOk;


    if(iMax<=0)return 0;

    pBuffer[0]=0;
    iLen=strlen(pPath);
    if(iLen+1>=iMax)return 0;

    if(pPath[0]!='/')
        {
        iOk=f_getcwd((TCHAR*)pBuffer,iMax);
        if(iOk!=FR_OK)return 0;
        iPos=strlen(pBuffer);
        if(iPos+1+iLen>=iMax)return 0;
        pBuffer[iPos]='/';
        iPos++;
        memcpy(pBuffer+iPos,pPath,iLen+1);
        iLen+=iPos;
        }
    else{
        memcpy(pBuffer,pPath,iLen+1);
        iPos=1;
        }


    while(iPos<iLen)
        {
        if(iPos>0 && pBuffer[iPos]=='/')                // filter "//"
            {
            memmove(pBuffer+iPos,pBuffer+iPos+1,iLen-iPos);
            iLen-=1;
            continue;
            }
                                                        // filter "./"
        if(iPos>0 && pBuffer[iPos]=='.' && pBuffer[iPos+1]=='/')
            {
            memmove(pBuffer+iPos,pBuffer+iPos+2,iLen-iPos-1);
            iLen-=2;
            continue;
            }
                                                        // filter "."
        if(iPos>1 && pBuffer[iPos]=='.' && pBuffer[iPos+1]==0)
            {
            iPos--;
            pBuffer[iPos]=0;
            iLen=iPos;
            break;
            }

                                                        // filter "../"
        if(iPos>0 && pBuffer[iPos]=='.' && pBuffer[iPos+1]=='.' && pBuffer[iPos+2]=='/')
            {
            iSlash=iPos;

            for(iPos--;iPos>0;iPos--)
                {
                if(pBuffer[iPos-1]=='/')break;
                }

            iAdd = iSlash-iPos+3;

            memmove(pBuffer+iPos,pBuffer+iSlash+3,iLen-iSlash-2);
            iLen-=iAdd;
            continue;
            }
                                                        // filter ".."
        if(iPos>1 && pBuffer[iPos]=='.' && pBuffer[iPos+1]=='.' && pBuffer[iPos+2]==0)
            {
            for(iPos-=2;iPos>0;iPos--)
                {
                if(pBuffer[iPos]=='/')break;
                }

            if(iPos<=0)
                {
                pBuffer[0]='/';
                iPos=1;
                }

            pBuffer[iPos]=0;
            iLen=iPos;
            break;
            }

        for(;pBuffer[iPos];iPos++)                      // to next slash
            {
            if(pBuffer[iPos]!='/')continue;
            iPos++;
            break;
            }
        }
return iLen;
}


//*****************************************************************************
//*
//*     SysOpenEx
//*
//*****************************************************************************
//  Opens a file
//  pFilename   : name of the file
//  uFlags      : acces flags
//                  SYS_READ
//                  SYS_WRITE
//                  SYS_CREATE
//                  SYS_TRUNC
//  pError      : here the error code (below zero or 0 if ok) will be stored.
//                if this pontier is zero no error code will be stored.
//  Return the file handle or 0 if an error occurs
SYS_API SysFile SysOpenEx(const char *pFilename,unsigned uFlags,int *pError)
{
FILINFO     sStat;
char            cTemp[MAX_TEMP_PATH];
FIL             iHandle;

SysDW           dwMode;
FRESULT             iOk;
SysUtc      uTime;
SysFile	hHandle;

    pFilename = SysCheckPath(pFilename,cTemp);
    dwMode    = (uFlags&SYS_WRITE)? FA_WRITE|FA_READ:FA_READ;
    if(uFlags&(SYS_TRUNC))dwMode|=FA_CREATE_ALWAYS|FA_WRITE|FA_READ;
    if(uFlags&SYS_CREATE )dwMode|=FA_CREATE_NEW|FA_WRITE|FA_READ;
    iOk=f_open(&iHandle,(TCHAR*)pFilename,(BYTE)dwMode);
    if(iOk!=FR_OK)
        {
		printf(" error %d ",iOk);
        if(pError)*pError=SysError(iOk);
        return 0;
        }
	
    
    hHandle=SYS_NEW_HANDLE(iHandle,pFilename,pError);

    if(uFlags&SYS_CREATE)
	{
		uTime = time(0);
		SysSetFileTime((SysFile)hHandle,uTime,uTime,uTime);	
	}
    if(uFlags&SYS_TRUNC) f_truncate(&iHandle);
    if(f_stat((TCHAR*)pFilename,&sStat)==FR_OK)
        {	
        if(sStat.fattrib&AM_DIR)
            {
            if(pError)*pError=SYS_ERR_ACCESS_DENIED;
            f_close(&iHandle);
            return 0;
            }
        }
	if(pError)*pError=SYS_ERR_NONE;
	
	
return hHandle;
}



//*****************************************************************************
//*
//*     SysClose
//*
//*****************************************************************************
//  close a file
//  return 0 if all is ok, or an error code below zero
SYS_API int SysClose(SysFile hHandle)
{
FRESULT     iOk;
FIL     *iFile;

    iFile = SYS_GET_HANDLE(hHandle);
    SYS_DEL_HANDLE(hHandle);

    iOk=f_close(iFile);
    if(iOk!=FR_OK)
        {
        return SysError(iOk);
        }	

return 0;
}

//*****************************************************************************
//*
//*     SysFlush
//*
//*****************************************************************************
//  flush the file buffers
//  return 0 if all is ok, or error code below zero
SYS_API int SysFlush(SysFile hHandle)
{
FRESULT     iOk;
FIL    *iFile;

    iFile = SYS_GET_HANDLE(hHandle);
    iOk= f_sync(iFile);
    if(iOk!=FR_OK)
        {
        return SysError(iOk);
        }
return 0;
}

//*****************************************************************************
//*
//*     SysSeek
//*
//*****************************************************************************
//  seeks the file pointer
//  return the new file position or an error code below zero
SYS_API int SysSeek(SysFile hHandle,SysDW dwOffset,int iType)
{
FIL     *iFile;
FRESULT iOk;

    iFile = SYS_GET_HANDLE(hHandle);

    iOk=f_lseek(iFile,(FSIZE_t)dwOffset);
    if(iOk!=FR_OK)
        {
        return SysError(iOk);
        }
	

return (int)f_tell(iFile);
}
//*****************************************************************************
//*
//*     SysSeekEx
//*
//*****************************************************************************
//  seeks the file pointer (64-bit-version)
//  return the new file position or an error code below zero
SYS_API SysPos SysSeekEx(SysFile hHandle,SysPos qwOffset,int iType)
{
FIL     *iFile;
SysPos  iPos;
FRESULT iOk;

    iFile = SYS_GET_HANDLE(hHandle);
	
	iOk=f_lseek(iFile,(FSIZE_t)qwOffset);
    if(iOk!=FR_OK)
    {
    return SysError(iOk);
    }

    /*iPos  = lseek64(iFile,qwOffset,iType);

    if(iPos<0)
        {
        return SysError();
        }*/


//return iPos;
return (SysPos)f_tell(iFile);
}
//*****************************************************************************
//*
//*     SysTell
//*
//*****************************************************************************
//  return the file position or an error code below zero
SYS_API int SysTell(SysFile hHandle)
{
FIL    *iFile;
    iFile = SYS_GET_HANDLE(hHandle);
return (int)f_tell(iFile);
}
//*****************************************************************************
//*
//*     SysTellEx
//*
//*****************************************************************************
//  return the file position or an error code below zero
SYS_API SysPos SysTellEx(SysFile hHandle)
{
FIL    *iFile;
SysPos  iPos;


    iFile = SYS_GET_HANDLE(hHandle);

    /*iPos  = lseek64(iFile,0,SEEK_CUR);

    if(iPos<0)
        {
        return SysError();
        }*/


return (SysPos)f_tell(iFile);
}
//*****************************************************************************
//*
//*     SysRead
//*
//*****************************************************************************
//  read data from a file
//  hHandle : file handle
//  pBuffer : buffer for read
//  dwSize  : byte count
//  return the count of read bytes or an error code below zero
SYS_API int SysRead(SysFile hHandle,void *pBuffer,SysDW dwSize)
{
int     iCount,i,j;
FIL     *iFile;
FRESULT iOk;
char	*pTemp;
int iSize=500;
int iSize1,iSize2;
	
	iSize1=dwSize/iSize; iSize2=dwSize%iSize;	
    iFile = SYS_GET_HANDLE(hHandle);
	
	pTemp=(char*)pBuffer;	
	for(i=0,j=0;i<iSize1;i++)
	{
	iOk =f_read(iFile,pTemp,iSize,(UINT*)&iCount);
	if(iOk!=FR_OK)
        {
        return SysError(iOk);
        }
	j+=iCount;
	pTemp+=iCount;
	}
	iOk =f_read(iFile,pTemp,iSize2,(UINT*)&iCount);
	if(iOk!=FR_OK)
        {
        return SysError(iOk);
        }
	j+=iCount;
return (UINT)j;
}

//*****************************************************************************
//*
//*     SysWrite
//*
//*****************************************************************************
//  write data to a file
//  hHandle : file handle
//  pBuffer : buffer to write
//  dwSize  : byte count
//  return the count of read bytes or an error code below zero
SYS_API int SysWrite(SysFile hHandle,const void *pBuffer,SysDW dwSize)
{
UINT     iCount,i,j;
FIL     *iFile;
FRESULT	iOk;
int iSize=500;
int iSize1,iSize2;

	
	iSize1=dwSize/iSize; iSize2=dwSize%iSize;

   iFile = SYS_GET_HANDLE(hHandle);
	
/*    iOk =f_write(iFile,pBuffer,(UINT)dwSize,&iCount);
	
    if(iOk!=FR_OK)
        {
        return SysError(iOk);
        }	
	f_read(iFile,)
return (int)iCount;

//return (int)dwSize;
*/
	char * pTemp=(char*)pBuffer;	
	for(i=0,j=0;i<iSize1;i++)
	{
	iOk =f_write(iFile,pTemp,iSize,(UINT*)&iCount);
	if(iOk!=FR_OK)
        {
        return SysError(iOk);
        }
	j+=iCount;
	pTemp+=iCount;
	}
	iOk =f_write(iFile,pTemp,iSize2,(UINT*)&iCount);
	if(iOk!=FR_OK)
        {
        return SysError(iOk);
        }
	j+=iCount;
return (UINT)j;

}


//*****************************************************************************
//*
//*     SysGetFileSize
//*
//*****************************************************************************
//  returns the file size or an error code below zero
SYS_API int SysGetFileSize(SysFile hHandle)
{
FIL     *iFile;
    iFile = SYS_GET_HANDLE(hHandle);
return f_size(iFile);
}

//*****************************************************************************
//*
//*     SysSemaphoreCreate
//*
//*****************************************************************************
//  creates a new semaphore, the has the value 1 after creation and isn't locked
//  returns a semaphore handle
SYS_API SysSemaphore SysSemaphoreCreate()
{
return xSemaphoreCreateMutex();
}

//*****************************************************************************
//*
//*     SysSemaphoreCreateEx
//*
//*****************************************************************************
//  creates a new semaphore with an init value
//	hSemaphore		: handle of the semaphore
//  iNumber 		: is the init value (>0 isn't locked)
//  returns a semaphore handle
SYS_API SysSemaphore SysSemaphoreCreateEx(int iNumber)
{
return xSemaphoreCreateCounting(50,iNumber);
}

//*****************************************************************************
//*
//*     SysSemaphoreDelete
//*
//*****************************************************************************
//  deletes a semaphore
//	hSemaphore		: handle of the semaphore
//  returns 1 if the semaphore was deleted or 0 if an error occurs
SYS_API void SysSemaphoreDelete(SysSemaphore hSemaphore)
{
vSemaphoreDelete( hSemaphore );
}


//*****************************************************************************
//*
//*     SysSemaphoreUnlock
//*
//*****************************************************************************
//  unlocks a sempahore, or increases the semaphore value if it was unlocked
//	hSemaphore		: handle of the semaphore
//  returns 1 or 0 if ann error occurs
SYS_API int SysSemaphoreUnlock(SysSemaphore hSemaphore)
{
return xSemaphoreGive( hSemaphore );
}


//*****************************************************************************
//*
//*     SysSemaphoreUnlockEx
//*
//*****************************************************************************
//  unlocks a sempahore, or increases the semaphore value several times
//	hSemaphore		: handle of the semaphore
//	iCount			: is the count of releases (must be greater 0)
//  returns 1 or 0 if ann error occurs
SYS_API int SysSemaphoreUnlockEx(SysSemaphore hSemaphore,int iCount)
{
if(!hSemaphore || iCount<=0)return 0;

	while(iCount>0)
		{
    		xSemaphoreGive( hSemaphore );
		iCount--;
		}

return 1;
}


//*****************************************************************************
//*
//*     SysSemaphoreLock
//*
//*****************************************************************************
//  waits until a sempahore is unlocket, or decreases the semaphore value
//  if it was unlocked
//  returns 1 or 0 if ann error occurs
SYS_API int SysSemaphoreLock(SysSemaphore hSemaphore)
{
if(!hSemaphore)return 0;
	
	xSemaphoreTake( hSemaphore,
                 (TickType_t) 0);


return 1;
}


//*****************************************************************************
//*
//*     SysSemaphoreLockWait
//*
//*****************************************************************************
//  waits some time until a sempahore is unlocket, or decreases the semaphore
//  value if it was unlocked.
//  dwMillisec  : is the maximum wait time in milliseconds
//  returns     0: Timeout
//              1: Unlock
SYS_API int SysSemaphoreLockWait(SysSemaphore hSemaphore,SysDW dwMillisec)
{
if(!hSemaphore)return 0;

return  xSemaphoreTake( hSemaphore, (TickType_t) (dwMillisec/portTICK_PERIOD_MS));
}

//*****************************************************************************
//*
//*		SysSemaphoreCount
//*
//*****************************************************************************
//	Gets the value of a semaphore
//	hSemaphore		: handle of the semaphore
//	Returns the semophore value between 0 and 0x70000000
SYS_API int SysSemaphoreCount(SysSemaphore hSemaphore)
{
SemData            *pSem;
int					iRet;


	if(!hSemaphore)return 0;

	   pSem=(SemData*)hSemaphore;
	if(pSem->iId!=0x1255AA78)return 0;						// check semophore

	   iRet = pSem->iValue;
	if(iRet<0)return 0;


return iRet;
return 0;
}


//*****************************************************************************
//*
//*     SysSleep
//*
//*****************************************************************************
//  suspents the current thread for some time
//  dwMillisec      : is the time in millisekonds
SYS_API void SysSleep(SysDW dwMillisec)
{
	vTaskDelay((TickType_t)(dwMillisec/portTICK_PERIOD_MS ));
}

//*****************************************************************************
//*
//*     SysThreadStart
//*
//*****************************************************************************
//  start a new thread
//  pStartProc  : function fpr the thread
//  dwStackSize : count of bytes for stack
//  pParam      : Parameter for pStartProc
//  iPriority   : -256 to 256 relative to the current priority or
//                    SYS_GLOBAL|iPriority for a global priority
//  pName       : taskname
//  returns the thread handle or 0 if an error occurs
SYS_API SysThread SysThreadStart(void( *pStartProc)(void*),SysDW dwStackSize,void *pParam,int iPriority,const char *pName)
{
TaskHandle_t  ulId=NULL;

    if(!dwStackSize)return 0;
    xTaskCreate(    pStartProc,
                            pName,
                            (unsigned short) dwStackSize,
                            pParam,
                            (UBaseType_t) iPriority,
                            &ulId
                          );
	
return (SysThread)ulId;
}


//*****************************************************************************
//*
//*     SysGetAttributes
//*
//*****************************************************************************
//  get the the attribute from a file
//  pFilename   : pointer to file name
//  return ns the attriputes or an error number below zero
//              	SYS_READ
//              	SYS_WRITE
//              	SYS_DIR
//              	SYS_HIDDEN
//              	SYS_ARCHIVE
SYS_API int SysGetAttributes(const char *pFilename)
{
char        cTemp[MAX_TEMP_PATH];
FILINFO sStat;
int         iRet;
FRESULT         iOk;
    pFilename = SysCheckPath((const char*)pFilename,cTemp);
    iOk = f_stat((const TCHAR*)pFilename,&sStat);
    if(iOk!=FR_OK)
        {
        return SysError(iOk);
        }
    iRet=(sStat.fattrib&AM_RDO)? SYS_READ:SYS_READ|SYS_WRITE;
    if(sStat.fattrib&AM_DIR)iRet|=SYS_DIR;
    if(sStat.fattrib&AM_HID)iRet|=SYS_HIDDEN;
    if(sStat.fattrib&AM_ARC)iRet|=SYS_ARCHIVE;
    if(sStat.fattrib&AM_SYS)iRet|=SYS_SYSTEM;
return iRet;
}


//*****************************************************************************
//*
//*     SysSetAttributes
//*
//*****************************************************************************
//  set the the attribute for a file
//  pFilename   : pointer to file name
// 	dwAttr		: are the new attributes
//              	SYS_READ
//              	SYS_WRITE
//              	SYS_HIDDEN
//              	SYS_ARCHIVE
//					SYS_DIR
//  returns null if ok or an error number below zero
//					SYS_ERR_??? (a value <0)
SYS_API int SysSetAttributes(const char *pFilename,SysDW dwAttr)
{
char        cTemp[MAX_TEMP_PATH];
int         iMode;
FRESULT         iOk;


    pFilename = SysCheckPath(pFilename,cTemp);

    if(dwAttr&(SYS_VOLUME|SYS_DIR))
        {
        return SYS_ERR_PARAM;
        }


   iMode = (dwAttr&SYS_WRITE)? 0:AM_RDO;
   if(dwAttr&SYS_ARCHIVE)iMode|=AM_ARC;
   if(dwAttr&SYS_SYSTEM )iMode|=AM_SYS;
   if(dwAttr&SYS_HIDDEN )iMode|=AM_HID;
    iOk = f_chmod((TCHAR*)pFilename,iMode,iMode);
    if(iOk!=FR_OK)
        {
        return SysError(iOk);
        }
return 0;
}



//*****************************************************************************
//*
//*     SysFindFirstEx
//*
//*****************************************************************************
//  search for a file
//  pFilename   : find path (with '*' and '?' )
//  pFind       : pointer to file structer
//  pError      : here the error code (below zero or 0 if ok) will be stored.
//                if this pontier is zero no error code will be stored.
//  returns a find handle or 0 if an error occurs
SYS_API SysFind SysFindFirstEx(const char *pFilename,SysFindStruct *pFind,int *pError)
{
FF_DIR     pDir;
char        cTemp[MAX_TEMP_PATH];
char        cPath[FIND_LEN+4];
SysFind     hHandle;
FILINFO sStat,sEntry;
unsigned    uNr;
int         iFlag;
int         iAttr;
int         iLen;
int         i;
FRESULT	iOk;
SysDW      uDosTime;

    if(!pFind)
        {
        if(pError)*pError=SYS_ERR_PARAM;
        return 0;
        }

    if(!iFindInit)                                  // init all
       {
       memset(aFindMem,0,sizeof(aFindMem));
       iFindInit=1;
       HANDLE_INIT();
       }
	HANDLE_LOCK();

    for(uNr=0;uNr<FIND_MAX;uNr++)                   // find empty handle
        {
        if(aFindMem[uNr].iFree==0)break;
        }

    if(uNr>=FIND_MAX)
        {
        HANDLE_UNLOCK();
        SysClearFindStruct(pFind);
        if(pError)*pError=SYS_ERR_NO_HANDLES;
        return 0;
        }

    aFindMem[uNr].iFree=1;

    HANDLE_UNLOCK();

//*****************************************************************************
    iLen=strlen(pFilename);
    for(i=iLen,iFlag=0;i>0;i--)						// search for end of path
        {
        if(pFilename[i]=='/' || pFilename[i]=='\\')
			{
			if(i+1<iLen)break;
            if(pError)*pError=SYS_ERR_FILE_NOT_FOUND;
            SysClearFindStruct(pFind);
        	aFindMem[uNr].iFree=0;
			return 0;
			}
        if(pFilename[i]!='*' && pFilename[i]!='?')	// is it a wildchar
			{
			continue;
			}
        iFlag=1;
        }


    if(!iFlag)                                      // no wildchars
        {
		pFilename = SysCheckPath(pFilename,cTemp);
		iLen=strlen(pFilename);		
		for(i=iLen;i>0;i--)
		{
			if(pFilename[i]=='/') break;
		}
		memcpy(aFindMem[uNr].cName,pFilename,i);
		aFindMem[uNr].cName[i]=0;
		if(i==(iLen-1)) 
		{
			aFindMem[uNr].cFind[0]='*';		//name ""
			aFindMem[uNr].cFind[1]=0;
		}
		else{ 
			iLen-=(i+1);
			memcpy(aFindMem[uNr].cFind,pFilename+i+1,iLen+1);
			aFindMem[uNr].cFind[iLen]=0;
		}

		iOk=f_findfirst(&pDir,&sEntry,(const TCHAR*)aFindMem[uNr].cName,(const TCHAR*)aFindMem[uNr].cFind);
		if(iOk==FR_OK&&sEntry.fname[0])
		{
			f_closedir(&pDir);
			aFindMem[uNr].iFree=0;

			strcpy((char*)pFind->cName,(const char*)sEntry.fname);
        	#if SYS_HAS_SHORT_NAME
       		SysCopyShortName(pFind->cShortName,pFind->cName);
       		#endif
        	pFind->dwSize       = (SysDW)sEntry.fsize;
        	uDosTime=sEntry.ftime | (sEntry.fdate<<16);
			SysTimeConvertDS(*(int*)&uDosTime,&pFind->dwTimeCreate);
			pFind->dwTimeWrite=pFind->dwTimeCreate;
			pFind->dwTimeAccess=pFind->dwTimeCreate;

	        iAttr=(sEntry.fattrib&AM_RDO)? SYS_READ:SYS_READ|SYS_WRITE;
			if(sEntry.fattrib&AM_ARC)iAttr|=SYS_ARCHIVE;
     	    if(sEntry.fattrib&AM_HID )iAttr|=SYS_HIDDEN;
     	    if(sEntry.fattrib&AM_SYS )iAttr|=SYS_SYSTEM;
    		if(sEntry.fattrib&AM_DIR)iAttr|=SYS_DIR;

			pFind->dwAttr=iAttr;
			hHandle=(SysFind)(FIND_ONE+1);
			
			if(pError)*pError=SYS_ERR_NONE;
			
			return hHandle;
		}
		
		if (iOk!=FR_OK||sEntry.fname[0]==0)
		{
			f_closedir(&pDir);
			aFindMem[uNr].iFree=0;
			if(pError)*pError=SYS_ERR_FILE_NOT_FOUND;
			hHandle=0;
			return hHandle;
		}
		
       }

    if(i>=FIND_LEN)                                  // is the path too long ?
        {
        if(pError)*pError=SYS_ERR_INVALID_FILENAME;
        SysClearFindStruct(pFind);
        aFindMem[uNr].iFree=0;
        return 0;
        }
    memcpy(cPath,pFilename,i);
    cPath[i]   = 0; //path
    iLen      -= i+1; //lenght of file name
    pFilename += i+1; //file name

    SysCheckPath(cPath,cTemp);                      // new path is in cTemp

    i = strlen(cTemp); //lenght of path
    if(i+iLen+2>=(int)sizeof(cTemp))
        {
        if(pError)*pError=SYS_ERR_INVALID_FILENAME;
        SysClearFindStruct(pFind);
        aFindMem[uNr].iFree=0;
        return 0;
        }
    memcpy(cTemp+i+1,pFilename,iLen+1);
    pFilename = cTemp; //name of whole dir (include finding object name)
    cTemp[i]  = '/';
    iLen     += i+1;//lenght of whole dir


    memcpy(aFindMem[uNr].cName,pFilename,i);
           aFindMem[uNr].cName[i]=0; //name of base dir
           aFindMem[uNr].iLen =i;//lenght of base dir


    if(iLen-i>=FIND_LEN)                             // is the name too long ?
        {
        if(pError)*pError=SYS_ERR_INVALID_FILENAME;
        SysClearFindStruct(pFind);
        aFindMem[uNr].iFree=0;
        return 0;
        }


    if(pFilename[i]=='/')i++;

    memcpy(aFindMem[uNr].cFind,pFilename+i,iLen-i+1); //name of finding object
           aFindMem[uNr].iNum = iLen-i; //lenght of finding object name


    if(!aFindMem[uNr].cName[0])                     // "" to "."
        {
        aFindMem[uNr].cName[0] = '/';
        aFindMem[uNr].cName[1] = 0;
        aFindMem[uNr].iLen     = 1;
        }          

	iOk=f_findfirst(&pDir,&sEntry,(const TCHAR*)aFindMem[uNr].cName,(const TCHAR*)aFindMem[uNr].cFind);
	if((iOk!=FR_OK)||(sStat.fname[0]==0))
	{
	SysClearFindStruct(pFind);
    aFindMem[uNr].iFree=0;
    if(pError)*pError=SYS_ERR_FILE_NOT_FOUND;
	return 0;			
	}
		
    i=aFindMem[uNr].iLen;
    aFindMem[uNr].cName[i]='/';i++;
    aFindMem[uNr].cName[i]=0;
    aFindMem[uNr].iLen++;
    aFindMem[uNr].pDir=&pDir;

    hHandle = (SysFind)(uNr+1);

	strcpy((char*)pFind->cName,(const char*)sEntry.fname);
        #if SYS_HAS_SHORT_NAME
//        SysCopyShortName(pFind->cShortName,pFind->cName);
		strcpy(pFind->cShortName,(const char*)sEntry.altname);
        #endif

        pFind->dwSize       = (SysDW)sEntry.fsize;
        	uDosTime=sEntry.ftime | (sEntry.fdate<<16);
		SysTimeConvertDS(*(int*)&uDosTime,&pFind->dwTimeCreate);
		pFind->dwTimeWrite=pFind->dwTimeCreate;
		pFind->dwTimeAccess=pFind->dwTimeCreate;

	    iAttr=(sEntry.fattrib&AM_RDO)? SYS_READ:SYS_READ|SYS_WRITE;
		if(sEntry.fattrib&AM_ARC)iAttr|=SYS_ARCHIVE;
     	if(sEntry.fattrib&AM_HID )iAttr|=SYS_HIDDEN;
     	if(sEntry.fattrib&AM_SYS )iAttr|=SYS_SYSTEM;
    	if(sEntry.fattrib&AM_DIR)iAttr|=SYS_DIR;

		pFind->dwAttr=iAttr;
   	if(pError)*pError=SYS_ERR_NONE;
	
	return hHandle;
}


//*****************************************************************************
//*
//*     SysFindNext
//*
//*****************************************************************************
//  search for the next file
//  hHandle : find handle
//  pFind   : pointer to file structer
//  returns 0 if a file was found or an error code below zero
SYS_API int SysFindNext(SysFind hHandle,SysFindStruct *pFind)
{
unsigned        uNr;
FF_DIR            *pDir;
FILINFO   sEntry;
FRESULT iOk;
SysDW      uDosTime;
int	iAttr;

	if(!pFind)
		{
		return SYS_ERR_PARAM;
		}

       uNr = (int)hHandle-1;
    if(uNr>=FIND_MAX || aFindMem[uNr].iFree==0)
        {
        SysClearFindStruct(pFind);
        return (uNr==FIND_ONE)? SYS_ERR_FILE_NOT_FOUND:SYS_ERR_PARAM;
        }

    pDir=aFindMem[uNr].pDir;
	iOk=f_findnext(pDir,&sEntry);
	if((iOk!=FR_OK)||(sEntry.fname[0]==0))
            {
	    		aFindMem[uNr].iFree=0;
            	SysClearFindStruct(pFind);
            	return SYS_ERR_FILE_NOT_FOUND;
            }
	strcpy(pFind->cName,(const char*)sEntry.fname);
    #if SYS_HAS_SHORT_NAME
//    SysCopyShortName(pFind->cShortName,pFind->cName);
	strcpy(pFind->cShortName,(const char*)sEntry.altname);
    #endif

        	pFind->dwSize       = (SysDW)sEntry.fsize;
        	uDosTime=sEntry.ftime | (sEntry.fdate<<16);
		SysTimeConvertDS(*(int*)&uDosTime,&pFind->dwTimeCreate);
		pFind->dwTimeWrite=pFind->dwTimeCreate;
		pFind->dwTimeAccess=pFind->dwTimeCreate;

	        iAttr=(sEntry.fattrib&AM_RDO)? SYS_READ:SYS_READ|SYS_WRITE;
		if(sEntry.fattrib&AM_ARC)iAttr|=SYS_ARCHIVE;
     	        if(sEntry.fattrib&AM_HID )iAttr|=SYS_HIDDEN;
     	        if(sEntry.fattrib&AM_SYS )iAttr|=SYS_SYSTEM;
    		if(sEntry.fattrib&AM_DIR)iAttr|=SYS_DIR;

		pFind->dwAttr=iAttr;	
return 0;
}


//*****************************************************************************
//*
//*     SysFindClose
//*
//*****************************************************************************
SYS_API int SysFindClose(SysFind hHandle)
{
unsigned    uNr;

    uNr = (int)hHandle-1;

    if(uNr>=FIND_MAX)
        {
        return (uNr==FIND_ONE)? SYS_ERR_NONE:SYS_ERR_PARAM;
        }

    if(aFindMem[uNr].iFree==0)
        {
        return SYS_ERR_PARAM;
        }


    f_closedir(aFindMem[uNr].pDir);
    aFindMem[uNr].iFree=0;


return 0;
}




//*****************************************************************************
//*
//*     SysGetFileInfo
//*
//*****************************************************************************
//  search for the next file
//  hHandle : find handle
//  pFind   : pointer to file structer
//  returns 0 if a file was found or an error code below zero
SYS_API int SysGetFileInfo(SysFile hHandle,SysFindStruct *pFind)
{
int             iAttr;
FILINFO     sStat;
char            cTemp[MAX_TEMP_PATH];
const char		*pFilename;
SysDW uDosTime;
FRESULT	iOk;


    pFilename=SysCheckPath(SYS_GET_NAME(hHandle),cTemp);
	

	iOk=f_stat((const TCHAR*)pFilename,&sStat);
    if(iOk!=FR_OK)
        {
		SYS_DEL_HANDLE(hHandle);
        SysClearFindStruct(pFind);
        return SysError(iOk);
        }

	#if SYS_HAS_SHORT_NAME
	SysCopyShortName(pFind->cShortName,(const char*)sStat.fname);
    #endif

   	 iAttr=(sStat.fattrib&AM_RDO)? SYS_READ:SYS_READ|SYS_WRITE;
	if(sStat.fattrib&AM_ARC)iAttr|=SYS_ARCHIVE;
	if(sStat.fattrib&AM_HID )iAttr|=SYS_HIDDEN;
	if(sStat.fattrib&AM_SYS )iAttr|=SYS_SYSTEM;
	if(sStat.fattrib&AM_DIR)iAttr|=SYS_DIR;

	pFind->dwAttr=iAttr;
	pFind->dwSize=sStat.fsize;

	uDosTime=sStat.ftime | (sStat.fdate<<16);
	SysTimeConvertDS(*(int*)&uDosTime,&pFind->dwTimeCreate);
	pFind->dwTimeWrite=pFind->dwTimeCreate;
	pFind->dwTimeAccess=pFind->dwTimeCreate;

    	pFind->dwAttr=iAttr;

        #if SYS_HAS_NAME_INFO
        strncpy(pFind->cName,SYS_GET_NAME(hHandle),sizeof(pFind->cName));
        #else
        pFind->cName[0] = 0;
        #endif

	#if SYS_HAS_SHORT_NAME
    	SysCopyShortName(pFind->cShortName,pFind->cName);
    	#endif

	

return 0;
}



//*****************************************************************************
//*
//*     SysMkdir
//*
//*****************************************************************************
//  create a new directory
//  return 0 if all ok
SYS_API int SysMkdir(const char *pFilename)
{
char        cTemp[MAX_TEMP_PATH];
FRESULT iOk;
    pFilename = SysCheckPath(pFilename,cTemp);

    if((iOk=f_mkdir((const TCHAR*)pFilename))==FR_OK) return 0;
return SysError(iOk);
}


//*****************************************************************************
//*
//*     SysRmdir
//*
//*****************************************************************************
//  remove a new directory
//  return 0 if all ok
SYS_API int SysRmdir(const char *pFilename)
{
char        cTemp[MAX_TEMP_PATH];
FRESULT	iOk;
    pFilename = SysCheckPath(pFilename,cTemp);

    if((iOk=f_unlink((const TCHAR*)pFilename))==FR_OK)return 0;

return SysError(iOk);
}


//*****************************************************************************
//*
//*     SysDelete
//*
//*****************************************************************************
//  delets a file
//  return 0 if all ok
SYS_API int SysDelete(const char *pFilename)
{
char        cTemp[MAX_TEMP_PATH];
FRESULT	iOk;
    pFilename = SysCheckPath(pFilename,cTemp);

    if((iOk=f_unlink((const TCHAR*)pFilename))==FR_OK)return 0;

return SysError(iOk);
}
//*****************************************************************************
//*
//*     SysRename
//*
//*****************************************************************************
//  moves or renames a file
SYS_API int SysRename(const char *pFilename,const char *pNewFilename)
{
char        cTemp1[MAX_TEMP_PATH];
char        cTemp2[MAX_TEMP_PATH];
FRESULT	iOk;

    pFilename     = SysCheckPath(pFilename   ,cTemp1);
    pNewFilename  = SysCheckPath(pNewFilename,cTemp2);

if((iOk=f_rename((const TCHAR*)pFilename,(const TCHAR*)pNewFilename))==FR_OK)return 0;

return SysError(iOk);

    //if(rename(pFilename,pNewFilename)==0)return 0;


//return SysError();
}

//*****************************************************************************
//*
//*     SysMove
//*
//*****************************************************************************
//  moves or renames a file
SYS_API int SysMove(const char *pFilename,const char *pNewFilename)
{
char        cTemp1[MAX_TEMP_PATH];
char        cTemp2[MAX_TEMP_PATH];
FRESULT	iOk;

    pFilename     = SysCheckPath(pFilename   ,cTemp1);
    pNewFilename  = SysCheckPath(pNewFilename,cTemp2);

    if((iOk=f_rename((const TCHAR*)pFilename,(const TCHAR*)pNewFilename))==FR_OK)return 0;

return SysError(iOk);
}


//*****************************************************************************
//*
//*     SysGetFileTime
//*
//*****************************************************************************
//  get the all times of a opened file
SYS_API int SysGetFileTime(SysFile hHandle,SysDW *pWrite,SysDW *pCreate,SysDW *pAccess)
{
FILINFO  sStat;
const char     *iFile;
FRESULT     iOk;
char            cTemp[MAX_TEMP_PATH];
SysDW uDosTime;

       iFile = SysCheckPath(SYS_GET_NAME(hHandle),cTemp);

       iOk = f_stat((const TCHAR*)iFile,&sStat);
      if(iOk!=FR_OK)
        {
        return SysError(iOk);
        }
    uDosTime=sStat.ftime | (sStat.fdate<<16);
    SysTimeConvertDS(*(int*)&uDosTime,pWrite);
    *pCreate=*pWrite;
    *pAccess=*pWrite;

return 0;
}


//*****************************************************************************
//*
//*     SysSetFileTime
//*
//*****************************************************************************
//  get the all times of an opened file
SYS_API int SysSetFileTime(SysFile hHandle,SysDW dwWrite,SysDW dwCreate,SysDW dwAccess)
{
FILINFO     sStat;
const char             *iFile;
FRESULT             iOk;
struct tm      *pGlobal;
char            cTemp[MAX_TEMP_PATH];

    iFile = SysCheckPath(SYS_GET_NAME(hHandle),cTemp);

    iOk = f_stat((const TCHAR*)iFile,&sStat);
        if(iOk!=FR_OK)
            {
            return SysError(iOk);
            }
     pGlobal=gmtime((const time_t*)&dwWrite);
     sStat.fdate = (pGlobal->tm_year - 1980)<<9;
     sStat.fdate |= (pGlobal->tm_mon)<<5;
     sStat.fdate |= (pGlobal->tm_mday);
	sStat.ftime = (pGlobal->tm_hour)<<11;
	sStat.ftime |= (pGlobal->tm_min)<<5;
	sStat.ftime |= (pGlobal->tm_sec)/2;
return 0;
}



//*****************************************************************************
//*
//*     SysGetVolumeInfo
//*
//*****************************************************************************
//  get info from a disk
//  pFilename   : name of the disk eg. "C:\"
//  pInfo       : pointer to the info stucture
//  iCalcFree   : should the free space be calculated
//  return 0 or an error code below zero
SYS_API int SysGetVolumeInfo(const char *pFilename,SysVolumeInfo *pInfo,int iCalcFree)
{
char            cTemp[MAX_TEMP_PATH];
const char     *pType;
DWORD	nclst;
FATFS	*fatfs;
FRESULT	iOk;

    if(!pInfo)
        {
        return SYS_ERR_PARAM;
        }
    pFilename = SysCheckPath(pFilename,cTemp);

	iOk = f_getfree ((const TCHAR*)"0", &nclst, &fatfs);
    if(iOk != FR_OK)
        {
        memset(pInfo,0,sizeof(pInfo->cName));
        return SysError(iOk);
        }

    strcpy(pInfo->cFsType,"???");
    strcpy(pInfo->cName  ,"");

    pInfo->dwBytesPerSector = _MAX_SS;
    pInfo->dwSectorsPerUnit = fatfs->csize;
    pInfo->dwTotalUnits     =  fatfs->n_fatent-2;
    pInfo->dwFreeUnits      =  nclst;
    pInfo->dwSerialNumber   =  fatfs->id;

    if(!iCalcFree)pInfo->dwFreeUnits=0;

        switch(fatfs->fs_type)
            {
	case FS_FAT12 :pType="FAT12";  break;
	case FS_FAT16 :pType="FAT16";  break;
	case FS_FAT32 :pType="FAT32";  break;
	case FS_EXFAT :pType="EXFAT";  break;
	default:          pType="???";  break;
            }
        strcpy(pInfo->cFsType,pType);
        strcpy(pInfo->cName  ,"");

return 0;
}



//*****************************************************************************
//*
//*     SysSetFileSize
//*
//*****************************************************************************
SYS_API int SysSetFileSize(SysFile hHandle,SysDW dwSize)
{
FIL     *iFile;
FRESULT     iOk;


    iFile = SYS_GET_HANDLE(hHandle);

    iOk=f_truncate(iFile);
    if(iOk<0)return SysError(iOk);


return 0;
}


//*****************************************************************************
//*
//*     SysTickInit
//*
//*****************************************************************************

//*****************************************************************************
//*
//*     SysTickCount
//*
//*****************************************************************************
SYS_API SysDW SysTickCount()
{
return  xTaskGetTickCount()*portTICK_PERIOD_MS;
}

//******************************************************************************
//*
//*		HeapCreate
//*
//******************************************************************************
void *HeapCreate(unsigned dwOptions,unsigned dwInitialSize,unsigned dwMaximumSize)
{
return (void*)0x22557733;
}

//******************************************************************************
//*
//*		HeapDestroy
//*
//******************************************************************************
int HeapDestroy(void *hHeap)
{
return (((unsigned)hHeap)==0x22557733);
}

//*****************************************************************************
//*
//*      HeapAlloc
//*
//*****************************************************************************
void *HeapAlloc(void *hHeap,SysDW dwFlags,SysDW dwBytes)
{
return malloc(dwBytes);
}

//*****************************************************************************
//*
//*      HeapReAlloc
//*
//*****************************************************************************
void *HeapReAlloc(void *hHeap,SysDW dwFlags,void *pOldMem,SysDW dwBytes)
{
return realloc(pOldMem,dwBytes);
}

//*****************************************************************************
//*
//*     HeapFree
//*
//*****************************************************************************
int HeapFree(void *hHeap,SysDW dwFlags,void *pMem)
{
    free(pMem);

return 1;
}

//*****************************************************************************
//*
//*      HeapAllocD
//*
//*****************************************************************************
#ifdef  HEAP_DEBUG
#define      MAX_ALLOCS 0x1000
char         cHeapNames [MAX_ALLOCS][28];
void        *pHeapPoints[MAX_ALLOCS];
unsigned     uHeapSizes [MAX_ALLOCS];
SysSemaphore hHeapSemaphore=0;

int dwHeapCount=0;
int dwHeapSize=0;
int dwHeapMax=0;

void   *HeapAllocD(void *hHeap,SysDW dwFlags,SysDW dwBytes,int iLine,const char *pFile)
{
void    *pMem;
char    *pData;
int      i,iLen;


    if(!hHeapSemaphore)
        {
        hHeapSemaphore=RTKCreateSemaphore(ST_COUNTING,1,NULL);
        }

        pMem=malloc(dwBytes);
    if(!pMem)return 0;

    SysSemaphoreLock(hHeapSemaphore);
    dwHeapCount++;
    for(i=0;i<MAX_ALLOCS;i++)if(!pHeapPoints[i])break;
    if(i<MAX_ALLOCS){pHeapPoints[i]=pMem;}
    SysSemaphoreUnlock(hHeapSemaphore);

    if(i<MAX_ALLOCS)
        {
        uHeapSizes[i]=dwBytes;
        pData=cHeapNames[i];
        SysSemaphoreLock(hHeapSemaphore);
        dwHeapSize+=dwBytes;
        SysSemaphoreUnlock(hHeapSemaphore);

        i=strlen(pFile);
        for(;i>0;i--)
            {
            if(pFile[i]=='\\'){i++;break;}
            if(pFile[i]=='/' ){i++;break;}
            }

           iLen = strlen(pFile+i);
        if(iLen>20)iLen=20;

        memcpy (pData,pFile+i,iLen);
        sprintf(pData+iLen,"(%i)",iLine);
        pData[27]=0;
        }

    if(dwHeapSize>dwHeapMax)dwHeapMax=dwHeapSize;


return pMem;
}


//*****************************************************************************
//*
//*      HeapReAllocD
//*
//*****************************************************************************
void *HeapReAllocD(void *hHeap,SysDW dwFlags,void *pOldMem,SysDW dwBytes,int iLine,const char *pFile)
{
void    *pMem;
char    *pData;
int      i,iLen;


    if(!pOldMem)return 0;

    if(!hHeapSemaphore)
        {
        hHeapSemaphore=SysSemaphoreCreate();
        }

        pMem=realloc(pOldMem,dwBytes);
    if(!pMem)return 0;

    SysSemaphoreLock(hHeapSemaphore);
    for(i=0;i<MAX_ALLOCS;i++)if(pHeapPoints[i]==pOldMem)break;
    if(i<MAX_ALLOCS){pHeapPoints[i]=pMem;}
    SysSemaphoreUnlock(hHeapSemaphore);

    if(i<MAX_ALLOCS)
        {
        SysSemaphoreLock(hHeapSemaphore);
        dwHeapSize-=uHeapSizes[i];
        dwHeapSize+=dwBytes;
        SysSemaphoreUnlock(hHeapSemaphore);
        uHeapSizes[i]=dwBytes;
        pData=cHeapNames[i];

        i=strlen(pFile);
        for(;i>0;i--)
            {
            if(pFile[i]=='\\'){i++;break;}
            if(pFile[i]=='/' ){i++;break;}
            }


           iLen = strlen(pFile+i);
        if(iLen>20)iLen=20;

        memcpy (pData,pFile+i,iLen);
        sprintf(pData+iLen,"(%i)",iLine);
        pData[27]=0;
        }

    if(dwHeapSize>dwHeapMax)dwHeapMax=dwHeapSize;


return pMem;
}

//*****************************************************************************
//*
//*      HeapFreeD
//*
//*****************************************************************************
int HeapFreeD(void *hHeap,SysDW dwFlags,void *pMem,int iLine,const char *pFile)
{
char    *pData=(char*)pMem;
int      i;


    if(!pMem)
        return 0;

    SysSemaphoreLock(hHeapSemaphore);
    dwHeapCount--;
    SysSemaphoreLock(hHeapSemaphore);

    for(i=0;i<MAX_ALLOCS;i++)if(pHeapPoints[i]==pMem)break;
    if(i<MAX_ALLOCS)
        {
        SysSemaphoreLock(hHeapSemaphore);
        dwHeapSize-=uHeapSizes[i];
        SysSemaphoreUnlock(hHeapSemaphore);
        cHeapNames [i][0]=0;
        pHeapPoints[i]=0;
        uHeapSizes [i]=0;
        }


    free(pMem);


return 1;
}

//*****************************************************************************
//*
//*      HeapPrint
//*
//*****************************************************************************
int HeapPrint()
{
char    *pData;
int      i;


    SysSleep(500);

    #ifdef NO_PRINT
    printf("\nHEAP COUNT = %i   size=%i  max=%i",dwHeapCount,dwHeapSize,dwHeapMax);
    #else
    nprintf("\nHEAP COUNT = %i   size=%i  max=%i",dwHeapCount,dwHeapSize,dwHeapMax);
    #endif

    for(i=0;i<MAX_ALLOCS;i++)
        {
        if(!pHeapPoints[i])continue;
        pData=cHeapNames[i];
        #ifdef NO_PRINT
        printf("\n   %08X --> \"%s\" = %i",pHeapPoints[i],pData,uHeapSizes[i]);
        #else
        nprintf("\n   %08X --> \"%s\" = %i",pHeapPoints[i],pData,uHeapSizes[i]);
        #endif
        }

    SysSleep(300);


return dwHeapCount;
}
#endif




